from snmp.exception import *
from snmp.message import ProtocolVersion
from snmp.pdu import ReportPDU, ResponsePDU
from snmp.pipeline import *
from snmp.scheduler import Scheduler
from snmp.security.levels import noAuthNoPriv
from snmp.security.usm import *
from snmp.transport import *
from snmp.transport.udp import *
from snmp.v1.manager import *
from snmp.v1.requests import *
from snmp.v2c.manager import *
from snmp.v2c.requests import *
from snmp.v3.interpreter import *
from snmp.v3.manager import *

class NoDefaultUser(SNMPException):
    pass

class Engine:
    TRANSPORTS = {
        cls.DOMAIN: cls for cls in [
            UdpIPv4Socket,
            UdpIPv6Socket,
        ]
    }

    def __init__(self,
        defaultVersion = ProtocolVersion.SNMPv3,
        defaultDomain = TransportDomain.UDP_IPv4,
        defaultCommunity = b"public",
        autowait = True,
        verboseLogging = False,
    ):
        self.defaultVersion         = defaultVersion
        self.defaultDomain          = defaultDomain
        self.defaultCommunity       = defaultCommunity
        self.autowaitDefault        = autowait

        self.multiplexor = UdpMultiplexor()
        self.scheduler = Scheduler(self.multiplexor.poll)

        self.v1_admin = SNMPv1RequestAdmin(self.scheduler)
        self.v2c_admin = SNMPv2cRequestAdmin(self.scheduler)

        self.usm = UserBasedSecurityModule()
        self.v3_sorter = SNMPv3MessageSorter(SNMPv3Interpreter(self.usm))
        self.v3_table = SNMPv3MessageTable()
        self.v3_sorter.register(ReportPDU, self.v3_table)
        self.v3_sorter.register(ResponsePDU, self.v3_table)

        self.decoder = VersionDecoder()
        self.pipeline = Catcher(self.decoder, verbose=verboseLogging)
        self.decoder.register(ProtocolVersion.SNMPv1, self.v1_admin)
        self.decoder.register(ProtocolVersion.SNMPv2c, self.v2c_admin)
        self.decoder.register(ProtocolVersion.SNMPv3, self.v3_sorter)

        self.transports = {}

    def __del__(self):
        self.multiplexor.close()

    def addUser(self,
        user,
        namespace = "",
        default = None,
        authProtocol = None,
        privProtocol = None,
        authSecret = None,
        privSecret = None,
        secret = None,
        defaultSecurityLevel = None,
    ):
        self.usm.addUser(
            user.encode(),
            namespace,
            default,
            authProtocol,
            privProtocol,
            authSecret,
            privSecret,
            secret,
            defaultSecurityLevel,
        )

    def createChannel(self, domain, address, localAddress, mtu):
        try:
            transportClass = self.TRANSPORTS[domain]
        except KeyError as err:
            errmsg = f"Unsupported transport domain: {transportClass.DOMAIN}"
            raise ValueError(errmsg) from err

        address = transportClass.normalizeAddress(
            address,
            AddressUsage.LISTENER,
        )

        localAddress = transportClass.normalizeAddress(localAddress)

        try:
            transports = self.transports[domain]
        except KeyError:
            transports = {}
            self.transports[domain] = transports

        try:
            transport = transports[localAddress]
        except KeyError:
            if mtu is None:
                transport = transportClass(*localAddress)
            else:
                transport = transportClass(*localAddress, mtu=mtu)

            self.multiplexor.register(transport, self.pipeline)
            transports[localAddress] = transport

        return TransportChannel(transport, address)

    def v1Manager(self, channel, autowait, community = None):
        if community is None:
            community = self.defaultCommunity

        return SNMPv1Manager(
            self.v1_admin,
            channel,
            community,
            autowait,
        )

    def v2cManager(self, channel, autowait, community = None):
        if community is None:
            community = self.defaultCommunity

        return SNMPv2cManager(
            self.v2c_admin,
            channel,
            community,
            autowait,
        )

    def v3Manager(self,
        channel,
        autowait,
        engineID = None,
        defaultSecurityLevel = None,
        defaultUser = None,
        namespace = "",
    ):
        if defaultUser is None:
            defaultUserName = self.usm.defaultUserName(namespace)

            if defaultUserName is None:
                errmsg = "An SNMPv3 Manager requires a default username." \
                    " Before calling Manager(), you should first configure" \
                    " the users "

                if namespace:
                    errmsg += f"in namespace \"{namespace}\""
                else:
                    errmsg += "for this engine"

                errmsg += " by calling addUser(). If you prefer to" \
                    " communicate unsecurely, you may instead pass a" \
                    " username to Manager() via the \"defaultUser\" keyword" \
                    " argument."

                raise TypeError(errmsg)
        else:
            defaultUserName = defaultUser.encode()

        if defaultSecurityLevel is None:
            defaultSecurityLevel = self.usm.defaultSecurityLevel(
                defaultUserName,
                namespace,
            )

            if defaultSecurityLevel is None:
                if defaultUser is None:
                    errmsg = "Successfully inferred the default username"

                    if namespace:
                        errmsg += f" for namespace \"{namespace}\""

                    errmsg += ", but not the default security level"
                    raise SNMPLibraryBug(errmsg)
                else:
                    defaultSecurityLevel = noAuthNoPriv
        else:
            maxSecurityLevel = self.usm.maxSecurityLevel(
                defaultUserName,
                namespace,
            )

            if defaultSecurityLevel > maxSecurityLevel:
                errmsg = "The security configuration for user" \
                    f" \"{defaultUserName.decode()}\""

                if namespace:
                    errmsg += f" in namespace \"{namespace}\""

                errmsg += " does not include a"

                if not maxSecurityLevel.auth:
                    errmsg += "n authentication protocol"
                elif defaultSecurityLevel.priv:
                    errmsg += " privacy protocol"

                errmsg += f"; please call the addUser() method to update" \
                    " the configuration."

                raise ValueError(errmsg)


        return SNMPv3Manager(
            self.scheduler,
            self.v3_table,
            self.v3_sorter,
            channel,
            namespace,
            defaultUserName,
            defaultSecurityLevel,
            engineID=engineID,
            autowait=autowait,
        )

    def Manager(self,
        address,
        version = None,
        domain = None,
        localAddress = None,
        mtu = None,
        autowait = None,
        **kwargs,
    ):
        if version is None:
            version = self.defaultVersion
        elif not isinstance(version, ProtocolVersion):
            version = ProtocolVersion(version)

        if domain is None:
            domain = self.defaultDomain

        channel = self.createChannel(domain, address, localAddress, mtu)

        if autowait is None:
            autowait = self.autowaitDefault

        if version == ProtocolVersion.SNMPv3:
            return self.v3Manager(channel, autowait, **kwargs)
        elif version == ProtocolVersion.SNMPv2c:
            return self.v2cManager(channel, autowait, **kwargs)
        elif version == ProtocolVersion.SNMPv1:
            return self.v1Manager(channel, autowait, **kwargs)
        else:
            raise ValueError(f"Unsupported protocol version: {str(version)}")



__all__ = ["ParseError", "Tag", "decode", "decodeExact", "encode"]

from enum import IntEnum

from snmp.exception import *
from snmp.utils import *

class ParseError(IncomingMessageErrorWithPointer):
    pass

class Tag:
    """Represents an ASN.1 BER tag."""
    class Class(IntEnum):
        """Named constants for the class bits of an ASN.1 BER tag."""
        UNIVERSAL         = 0
        APPLICATION       = 1
        CONTEXT_SPECIFIC  = 2
        PRIVATE           = 3

    def __init__(self, number, constructed = False, cls = Class.UNIVERSAL):
        self.cls = cls
        self.constructed = constructed
        self.number = number

    def __eq__(self, other):
        if other.__class__ != self.__class__:
            return NotImplemented

        return (    # type: ignore[no-any-return]
            self.cls == other.cls
        and self.constructed == other.constructed
        and self.number == other.number
        )

    def __hash__(self):
        return hash((self.cls, self.constructed, self.number))

    def __repr__(self):
        return f"Tag({self.number}, {self.constructed}, {self.cls})"

    @classmethod
    def decode(cls, data):
        """Extract the tag from an ASN.1 BER string.

        This function decodes the tag portion of a BER string an returns it as
        a Tag object. It also returns a subbytes of the data argument
        referencing everything after the tag.
        """
        try:
            byte, ptr = data.pop_front()
        except IndexError as err:
            raise ParseError("Missing tag", data) from err

        class_      = (byte & 0xc0) >> 6
        constructed = (byte & 0x20) != 0
        number      = (byte & 0x1f)

        if number == 0x1f:
            number = 0
            byte = 0x80
            while byte & 0x80:
                try:
                    byte, ptr = ptr.pop_front()
                except IndexError as err:
                    raise ParseError("Incomplete tag", data) from err

                number <<= 7
                number |= byte & 0x7f

        return cls(number, constructed, cls.Class(class_)), ptr

    def encode(self):
        """Encode a Tag under ASN.1 Basic Encoding Rules."""
        byte = (self.cls << 6) & 0xc0

        if self.constructed:
            byte |= 0x20

        arr = bytearray()
        if self.number < 0x1f:
            byte |= self.number
        else:
            byte |= 0x1f

            number = self.number
            indicator = 0
            while number:
                arr.append(indicator | (number & 0x7f))
                indicator = 0x80
                number >>= 7

        arr.append(byte)
        return bytes(reversed(arr))

def decode_length(data):
    """Decode the length field of an ASN.1 BER string.

    The provided data argument should contain most of a BER string, starting
    after the tag. This function will decode the length field and return it as
    an int, along with a new subbytes object referencing to the portion of the
    data argument immediately following the length field.
    """
    try:
        length, ptr = data.pop_front()
    except IndexError as err:
        raise ParseError("Missing length", data) from err

    if length & 0x80:
        n = length & 0x7f

        length = 0
        for i in range(n):
            try:
                byte, ptr = ptr.pop_front()
            except IndexError as err:
                raise ParseError("Incomplete length", data) from err

            length <<= 8
            length |= byte

    return length, ptr

def encode_length(length):
    """Encode the length of a message under ASN.1 Basic Encoding Rules."""
    if length < 0x80:
        return bytes([length])

    arr = bytearray()
    while length:
        arr.append(length & 0xff)
        length >>= 8

    if len(arr) > 0x7f:
        raise ValueError("Length too large for definite-length encoding")

    arr.append(0x80 | len(arr))
    return bytes(reversed(arr))

def decode(data):
    original = subbytes(data)
    tag, ptr = Tag.decode(original)
    length, ptr = decode_length(ptr)

    if len(ptr) < length:
        raise ParseError("Incomplete value", original)

    body, tail = ptr.split(length)
    return tag, body, tail

def decodeExact(data):
    tag, body, tail = decode(data)

    if tail:
        raise ParseError(f"Trailing bytes", tail)

    return tag, body

def encode(tag, data):
    """Encode a message under ASN.1 Basic Encoding Rules."""
    return tag.encode() + encode_length(len(data)) + data



__all__ = ["AuthProtocol", "PrivProtocol", "UserBasedSecurityModule"]

class AuthProtocol:
    def __init__(self, key):
        raise NotImplementedError()

    def __eq__(self, other):
        return NotImplemented

    @classmethod
    def computeKey(cls, secret):
        raise NotImplementedError()

    @classmethod
    def localizeKey(cls, key, engineID):
        raise NotImplementedError()

    @classmethod
    def localize(cls, secret, engineID):
        return cls.localizeKey(cls.computeKey(secret), engineID)

    @property
    def msgAuthenticationParameters(self):
        raise NotImplementedError()

    def sign(self, data):
        raise NotImplementedError()

class PrivProtocol:
    def __init__(self, key):
        raise NotImplementedError()

    def decrypt(self, data, engineBoots, engineTime, salt):
        raise NotImplementedError()

    def encrypt(self, data, engineBoots, engineTime):
        raise NotImplementedError()

from .implementation import *
from .users import *



from snmp.security.usm import PrivProtocol
from snmp.typing import *

class AesCfb128(PrivProtocol):
    def __init__(self, key: bytes) -> None:
        ...

    def decrypt(self,
        data: bytes,
        engineBoots: int,
        engineTime: int,
        salt: bytes,
    ) -> bytes:
        ...

    def encrypt(self,
        data: bytes,
        engineBoots: int,
        engineTime: int,
    ) -> Tuple[bytes, bytes]:
        ...

class DesCbc(PrivProtocol):
    def __init__(self, key: bytes) -> None:
        ...

    def decrypt(self,
        data: bytes,
        engineBoots: int,
        engineTime: int,
        salt: bytes,
    ) -> bytes:
        ...

    def encrypt(self,
        data: bytes,
        engineBoots: int,
        engineTime: int,
    ) -> Tuple[bytes, bytes]:
        ...



__all__ = ["NumberAuthority", "NumberGenerator"]

from random import randint

class NumberGenerator:
    """Generate integers spanning a specific range.

    Given an integer size of n bits, an instance of this class will generate
    a sequence of length 2^n, where each n-bit integer appears exactly once.
    The first 2^n-1 numbers in the sequence are guaranteed to be non-zero; or,
    in other words, the last number in the sequence is always 0. After 2^n
    iterations, the sequence repeats from the beginning.

    The signed argument determines whether the generated numbers should use
    two's-complement encoding (i.e. cover the range -(2^(n-1)) to 2^(n-1)-1,
    inclusive), or unsigned encoding (0 to 2^n-1).
    """

    def __init__(self, n, signed = True):
        half = 1 << (n-1)

        self.previous = 0
        self.range = half << 1
        self.step = 2 * randint(1, half) - 1
        self.wrap = self.range - 1

        if signed:
            self.wrap -= half

    def __iter__(self):
        return self

    def __next__(self):
        """Return the next number in the sequence."""
        self.previous += self.step

        if self.previous > self.wrap:
            self.previous -= self.range

        return self.previous

class NumberAuthority:
    def __init__(self, attempts=10):
        self.attempts = attempts
        self.generator = self.newGenerator()
        self.reserved = set()

    def release(self, number):
        try:
            self.reserved.remove(number)
        except KeyError as err:
            raise self.DeallocationFailure(number) from err

    def reserve(self):
        for attempt in range(self.attempts):
            number = next(self.generator)

            if number == 0:
                self.generator = self.newGenerator()
                number = next(self.generator)

            if number not in self.reserved:
                self.reserved.add(number)
                return number

        raise self.AllocationFailure(self.attempts)



__all__ = ["UserBasedSecurityModule"]

from time import time

from snmp.exception import *
from snmp.pdu import *
from snmp.security import *
from snmp.security.levels import *
from snmp.security.models import *
from snmp.smi import *
from snmp.v3.message import *

from .credentials import *
from .parameters import *
from .stats import *
from .timekeeper import *
from .users import *

class UsmUnsupportedSecLevel(IncomingMessageError):
    pass

class UsmUnknownEngineID(IncomingMessageError):
    pass

class UsmUnknownUserName(IncomingMessageError):
    pass

class UsmWrongDigest(IncomingMessageError):
    pass

class UsmDecryptionError(IncomingMessageError):
    pass

class UserBasedSecurityModule:
    MODEL = SecurityModel.USM

    def __init__(self, namespace = None, engineID = None, engineBoots = 0):
        self.timekeeper = TimeKeeper()
        self.users = UserRegistry()

        if engineID is not None:
            if namespace is None:
                raise TypeError("missing the required 'namespace' argument")

            now = time()
            self.engineTime = EngineTime(now, engineBoots, authoritative=True)
            self.namespace = namespace

        self.engineID = engineID

        self.unsupportedSecLevels   = 0
        self.notInTimeWindows       = 0
        self.unknownUserNames       = 0
        self.unknownEngineIDs       = 0
        self.wrongDigests           = 0
        self.decryptionErrors       = 0

    def addUser(self,
        userName,
        namespace,
        default = None,
        authProtocol = None,
        privProtocol = None,
        authSecret = None,
        privSecret = None,
        secret = None,
        defaultSecurityLevel = None,
    ):
        self.users.addUser(
            userName,
            namespace,
            default,
            authProtocol,
            privProtocol,
            authSecret,
            privSecret,
            secret,
            defaultSecurityLevel,
        )

    def defaultSecurityLevel(self, userName, namespace):
        try:
            return self.users.defaultSecurityLevel(userName, namespace)
        except KeyError:
            return None

    def defaultUserName(self, namespace):
        try:
            return self.users.defaultUserName(namespace)
        except KeyError:
            return None

    def maxSecurityLevel(self, userName, namespace):
        try:
            return self.users.defaultSecurityLevel(userName, namespace)
        except KeyError:
            return noAuthNoPriv

    ### Methods for outgoing messages

    def outgoingNamespace(self, message):
        if message.securityEngineID == self.engineID:
            return self.namespace
        else:
            try:
                return next(iter(message.securityName.namespaces))
            except StopIteration as err:
                userName = message.securityName.userName
                errmsg = f"No namespace given for user {userName}"
                raise TypeError(errmsg) from err

    def outgoingUser(self, message):
        return self.users.credentials(
            message.securityName.userName,
            self.outgoingNamespace(message),
            message.securityEngineID,
        )

    def outgoingTime(self, message, timestamp):
        engineID = message.securityEngineID
        if engineID == self.engineID:
            snmpEngineBoots = self.engineTime.snmpEngineBoots
            snmpEngineTime = self.engineTime.snmpEngineTime(timestamp)
            return snmpEngineBoots, snmpEngineTime
        elif message.header.flags.authFlag:
            return self.timekeeper.getEngineTime(engineID, timestamp)
        else:
            return 0, 0

    def prepareOutgoing(self, message, timestamp = None):
        if timestamp is None:
            timestamp = time()

        scopedPduData = message.scopedPDU
        placeholder = b""
        salt = b""

        engineBoots, engineTime = self.outgoingTime(message, timestamp)

        if message.header.flags.authFlag:
            user = self.outgoingUser(message)
            placeholder = user.signaturePlaceholder()

            if message.header.flags.privFlag:
                scopedPduData, salt = user.encrypt(
                    message.scopedPDU,
                    engineBoots,
                    engineTime,
                )

        securityParameters = UnsignedUsmParameters(
            message.securityEngineID,
            engineBoots,
            engineTime,
            message.securityName.userName,
            placeholder,
            salt,
        )

        wireMessage = SNMPv3WireMessage(
            message.header,
            scopedPduData,
            OctetString(securityParameters.encode())
        )

        if message.header.flags.authFlag:
            return user.sign(wireMessage)
        else:
            return wireMessage.encode()

    ### Methods for incoming messages

    def candidateNamespaces(self, sp):
        if self.engineID is not None and sp.engineID == self.engineID:
            if self.users.exists(sp.userName, self.namespace):
                return [self.namespace]
            else:
                return []
        else:
            return list(self.users.namespaces(sp.userName))

    def authenticate(self, securityParameters, namespaces):
        authMatch = None
        authEnabled = False
        authenticated = list()

        if not namespaces:
            raise UsmUnknownUserName(securityParameters.userName)

        for namespace in namespaces:
            user = self.users.credentials(
                securityParameters.userName,
                namespace,
                securityParameters.engineID,
            )

            if authMatch is None:
                try:
                    user.verifySignature(securityParameters.signature)
                except AuthenticationNotEnabled:
                    pass
                except InvalidSignature:
                    authEnabled = True
                else:
                    authMatch = user.withoutPrivacy()
                    authenticated.append(namespace)
            elif user.withoutPrivacy() == authMatch:
                authenticated.append(namespace)

        if authenticated:
            return authenticated
        elif authEnabled:
            raise UsmWrongDigest(securityParameters.signature)
        else:
            raise UsmUnsupportedSecLevel(authNoPriv)

    def decrypt(self, encryptedPDU, securityParameters, namespaces):
        privMatch = None
        privEnabled = False
        successful = list()

        for namespace in namespaces:
            user = self.users.credentials(
                securityParameters.userName,
                namespace,
                securityParameters.engineID,
            )

            if privMatch is None:
                try:
                    scopedPDU = user.decrypt(
                        encryptedPDU,
                        securityParameters.engineBoots,
                        securityParameters.engineTime,
                        securityParameters.salt,
                    )
                except PrivacyNotEnabled:
                    pass
                except Exception:
                    privEnabled = True
                else:
                    privMatch = user
                    successful.append(namespace)
            elif user == privMatch:
                successful.append(namespace)

        if successful:
            return scopedPDU, successful
        elif privEnabled:
            raise UsmDecryptionError(encryptedPDU)
        else:
            raise UsmUnsupportedSecLevel(authPriv)

    def verifyIncomingTime(self, message, sp, timestamp):
        if sp.engineID == self.engineID:
            if message.header.flags.authFlag:
                self.engineTime.verifyTimeliness(
                    timestamp,
                    sp.engineBoots,
                    sp.engineTime,
                )
            else:
                pass
        else:
            if message.header.flags.authFlag:
                self.timekeeper.updateAndVerify(
                    sp.engineID,
                    timestamp,
                    sp.engineBoots,
                    sp.engineTime,
                )
            else:
                self.timekeeper.hint(
                    sp.engineID,
                    timestamp,
                    sp.engineBoots,
                    sp.engineTime,
                )

    def processIncoming(self, message, timestamp = None):
        if timestamp is None:
            timestamp = time()

        sp_original = message.securityParameters.original
        sp = SignedUsmParameters.decodeExact(sp_original)

        try:
            reportSecurityLevel = noAuthNoPriv

            if message.header.flags.privFlag:
                reportable = message.header.flags.reportableFlag
                requestID = 0
                contextName = b""
            else:
                scopedPDU = message.scopedPduData
                reportable = scopedPDU.pdu.CONFIRMED_CLASS
                requestID = scopedPDU.pdu.requestID
                contextName = scopedPDU.contextName

            if reportable and sp.engineID != self.engineID:
                raise UsmUnknownEngineID(sp.engineID)

            namespaces = self.candidateNamespaces(sp)

            if message.header.flags.authFlag:
                namespaces = self.authenticate(sp, namespaces)
                reportSecurityLevel = authNoPriv

                if message.header.flags.privFlag:
                    scopedPDU, namespaces = self.decrypt(
                        message.scopedPduData,
                        sp,
                        namespaces,
                    )

                    reportable = scopedPDU.pdu.CONFIRMED_CLASS
                    requestID = scopedPDU.pdu.requestID
                    contextName = scopedPDU.contextName

                    if reportable and sp.engineID != self.engineID:
                        raise UsmUnknownEngineID(sp.engineID)

            self.verifyIncomingTime(message, sp, timestamp)

        except UsmUnsupportedSecLevel as err:
            self.unsupportedSecLevels += 1
            if reportable:
                oid = usmStatsUnsupportedSecLevelsInstance
                value = Counter32(self.unsupportedSecLevels)
            else:
                raise
        except UsmNotInTimeWindow as err:
            self.notInTimeWindows += 1
            if reportable:
                oid = usmStatsNotInTimeWindowsInstance
                value = Counter32(self.notInTimeWindows)
            else:
                raise
        except UsmUnknownUserName as err:
            self.unknownUserNames += 1
            if reportable:
                oid = usmStatsUnknownUserNamesInstance
                value = Counter32(self.unknownUserNames)
            else:
                raise
        except UsmUnknownEngineID as err:
            self.unknownEngineIDs += 1
            if reportable and self.engineID is not None:
                oid = usmStatsUnknownEngineIDsInstance
                value = Counter32(self.unknownEngineIDs)
            else:
                raise
        except UsmWrongDigest as err:
            self.wrongDigests += 1
            if reportable:
                oid = usmStatsWrongDigestsInstance
                value = Counter32(self.wrongDigests)
            else:
                raise
        except UsmDecryptionError as err:
            self.decryptionErrors += 1
            if reportable:
                oid = usmStatsDecryptionErrorsInstance
                value = Counter32(self.decryptionErrors)
            else:
                raise
        else:
            return SNMPv3Message(
                message.header,
                scopedPDU,
                sp.engineID,
                SecurityName(sp.userName, *namespaces),
            )

        raise ReportMessage(
            SNMPv3Message(
                HeaderData(
                    message.header.msgID,
                    message.header.maxSize,
                    MessageFlags(reportSecurityLevel),
                    self.MODEL,
                ),
                ScopedPDU(
                    ReportPDU(VarBind(oid, value), requestID=requestID),
                    self.engineID,
                    contextName,
                ),
                self.engineID,
                SecurityName(sp.userName, self.namespace),
            )
        )



import os

if os.name == "nt":
    libcrypto = "libcrypto_static"
else:
    libcrypto = "crypto"

import cffi # type: ignore[import-untyped]

ffi = cffi.FFI()
ffi.cdef("""
typedef struct evp_cipher_st EVP_CIPHER;
typedef struct evp_cipher_ctx_st EVP_CIPHER_CTX;

EVP_CIPHER_CTX *EVP_CIPHER_CTX_new(void);
void EVP_CIPHER_CTX_free(EVP_CIPHER_CTX *c);

int EVP_EncryptInit(EVP_CIPHER_CTX *ctx, const EVP_CIPHER *type,
                    const unsigned char *key, const unsigned char *iv);
int EVP_EncryptUpdate(EVP_CIPHER_CTX *ctx, unsigned char *out,
                      int *outl, const unsigned char *in, int inl);

int EVP_DecryptInit(EVP_CIPHER_CTX *ctx, const EVP_CIPHER *type,
                    const unsigned char *key, const unsigned char *iv);
int EVP_DecryptUpdate(EVP_CIPHER_CTX *ctx, unsigned char *out,
                      int *outl, const unsigned char *in, int inl);

const EVP_CIPHER *EVP_des_cbc(void);
const EVP_CIPHER *EVP_aes_128_cfb128(void);
""")

ffi.set_source(
    "snmp.openssl",
    "#include <openssl/evp.h>",
    libraries=[libcrypto],
)



__all__ = ["UserRegistry"]

from .credentials import *

class UserConfig:
    def __init__(self, credentials, defaultSecurityLevel):
        self.credentials = credentials
        self.defaultSecurityLevel = defaultSecurityLevel

    @property
    def maxSecurityLevel(self):
        return self.credentials.maxSecurityLevel

class NamespaceConfig:
    def __init__(self):
        self.defaultUserName = None
        self.users = {}

    def __contains__(self, userName):
        return userName in self.users

    def __getitem__(self, userName):
        return self.users[userName]

    def addUser(self, userName, credentials, defaultSecurityLevel, default):
        config = UserConfig(credentials, defaultSecurityLevel)
        self.users[userName] = config

        if default:
            self.defaultUserName = userName

class UserRegistry:
    def __init__(self):
        self.localizedCredentials = {}
        self.namespaceConfigs = {}

    def addUser(self,
        userName,
        namespace,
        default = None,
        authProtocol = None,
        privProtocol = None,
        authSecret = None,
        privSecret = None,
        secret = None,
        defaultSecurityLevel = None
    ):
        if not userName:
            raise ValueError(f"Empty userName")
        elif len(userName) > 32:
            raise ValueError(f"userName is too long: {userName!r}")

        newNamespace = namespace not in self.namespaceConfigs

        if default is None:
            default = newNamespace
        elif newNamespace and not default:
            msg = "default may not be False for the first user in a namespace"
            raise ValueError(msg)

        credentials = self.makeCredentials(
            authProtocol,
            privProtocol,
            authSecret,
            privSecret,
            secret,
        )

        if defaultSecurityLevel is None:
            defaultSecurityLevel = credentials.maxSecurityLevel
        elif defaultSecurityLevel > credentials.maxSecurityLevel:
            errmsg = f"Unable to support {defaultSecurityLevel}"
            if credentials.maxSecurityLevel.auth:
                errmsg += " without a privProtocol"
            else:
                errmsg += " without an authProtocol"

            raise ValueError(errmsg)

        if newNamespace:
            config = NamespaceConfig()
        else:
            config = self.namespaceConfigs[namespace]

        config.addUser(userName, credentials, defaultSecurityLevel, default)

        if newNamespace:
            self.namespaceConfigs[namespace] = config
            self.localizedCredentials[namespace] = {}

        self.localizedCredentials[namespace][userName] = {}

    def namespaces(self, userName):
        for name, config in self.namespaceConfigs.items():
            if userName in config:
                yield name

    def exists(self, userName, namespace):
        try:
            config = self.namespaceConfigs[namespace]
        except KeyError:
            return False

        return userName in config

    def credentials(self, userName, namespace, engineID):
        try:
            return self.localizedCredentials[namespace][userName][engineID]
        except KeyError:
            pass

        try:
            creds = self.namespaceConfigs[namespace][userName].credentials
        except KeyError as err:
            raise ValueError(
                f"User {userName!r} is not defined"
                f" in namespace \"{namespace}\""
            ) from err

        localized = creds.localize(engineID)
        self.localizedCredentials[namespace][userName][engineID] = localized
        return localized

    def defaultSecurityLevel(self, userName, namespace):
        return self.namespaceConfigs[namespace][userName].defaultSecurityLevel

    def defaultUserName(self, namespace):
        return self.namespaceConfigs[namespace].defaultUserName

    def maxSecurityLevel(self, userName, namespace):
        return self.namespaceConfigs[namespace][userName].maxSecurityLevel

    @staticmethod
    def makeCredentials(
        authProtocol,
        privProtocol,
        authSecret,
        privSecret,
        secret,
    ):
        if authProtocol is None:
            if (privProtocol is not None
            or authSecret is not None
            or secret is not None):
                raise TypeError("missing required argument: 'authProtocol'")
        elif authSecret is None and secret is None:
            raise TypeError("missing required argument: 'authSecret'")
        elif authSecret is not None and secret is not None:
            raise TypeError("'authSecret' and 'secret' are mutually exclusive")

        if privProtocol is None and privSecret is not None:
            raise TypeError("missing required argument: 'privProtocol'")

        if authProtocol is None:
            credentials = Credentials()
        elif privProtocol is None:
            credentials = AuthCredentials(authProtocol, authSecret or secret)
        else:
            credentials = AuthPrivCredentials(
                authProtocol,
                privProtocol,
                authSecret,
                privSecret,
                secret,
            )

        return credentials



__all__ = ["BadVersion", "ProtocolVersion", "VersionOnlyMessage"]

import enum

from snmp.exception import *
from snmp.smi import *
from snmp.utils import *

class BadVersion(IncomingMessageErrorWithPointer):
    pass

class ProtocolVersion(enum.IntEnum):
    SNMPv1  = 0
    SNMPv2c = 1
    SNMPv3  = 3

    # Python 3.11 changes IntEnum.__str__()
    __str__ = enum.Enum.__str__

class VersionOnlyMessage(Sequence):
    def __init__(self, version):
        self.version = version

    def __iter__(self):
        yield Integer(self.version)

    def __len__(self):
        return 1

    def __repr__(self):
        return f"{typename(self)}({str(self.version)})"

    @classmethod
    def deserialize(cls, data):
        msgVersion, tail = Integer.decode(data)

        try:
            version = ProtocolVersion(msgVersion.value)
        except ValueError as err:
            errmsg = f"Invalid msgVersion: {msgVersion.value}"
            raise BadVersion(errmsg, data, tail) from err

        return cls(version)



__all__ = [
    "AddressUsage", "TransportChannel",
    "TransportDomain", "TransportMultiplexor"
]

import enum
import os
import socket

AddressUsage = enum.Enum(
    "AddressUsage",
    ["LISTENER", "TRAP_LISTENER", "SENDER"]
)

class TransportDomain(enum.Enum):
    def __init__(self, family, loopback, default):
        self.address_family = family
        self.loopback_address = loopback
        self.default_address = default

    UDP_IPv4 = socket.AF_INET, "127.0.0.1", "0.0.0.0"
    UDP_IPv6 = socket.AF_INET6, "::1", "::"

class TransportChannel:
    def __init__(self, transport, address):
        self.transport = transport
        self.address = address

    @property
    def domain(self):
        return self.transport.DOMAIN

    @property
    def msgMaxSize(self):
        return self.transport.recvSize

    def send(self, data):
        self.transport.send(data, self.address)

class TransportMultiplexor:
    def register(self, sock, listener):
        raise NotImplementedError()

    def listen(self):
        done = False
        while not done:
            done = self.poll()

    def poll(self, timeout = None):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

supported = ("posix")
if os.name in supported:
    package = os.name
else:
    package = "generic"

package = "{}.{}".format(__name__, package)



__all__ = [
    "SNMPException",
    "SNMPLibraryBug",
    "IncomingMessageError",
    "IncomingMessageErrorWithPointer",
]

import os

from snmp.utils import subbytes

class SNMPException(Exception):
    """Base class for all run-time exceptions in this library."""

class SNMPLibraryBug(AssertionError):
    """Base class for logic errors in the code of this library."""

class IncomingMessageError(SNMPException):
    """An error indicating a received message is invalid in some way."""

class IncomingMessageErrorWithPointer(IncomingMessageError):
    def __init__(self, msg, data, tail = None):
        super().__init__(msg)

        if tail is None:
            self.data = subbytes(data)
        else:
            self.data = subbytes(data, stop=len(data) - len(tail))

    def __str__(self):
        return f"{super().__str__()}{os.linesep}{self.data}"



__all__ = ["SecurityModel", "UnknownSecurityModel"]

import enum

from snmp.exception import IncomingMessageErrorWithPointer

class UnknownSecurityModel(IncomingMessageErrorWithPointer):
    pass

class SecurityModel(enum.IntEnum):
    USM = 3



__all__ = [
    "RequestError",
    "AuthenticationFailure", "PrivacyFailure", "TimeWindowFailure",
    "UnknownUserName", "UnsupportedSecurityLevel", "UnhandledReport",
    "InvalidResponseField", "NamespaceMismatch",
    "SNMPv3Manager", "SNMPv3MessageTable",
]

import collections
import re
import weakref

from snmp.exception import *
from snmp.pdu import *
from snmp.scheduler import *
from snmp.security import *
from snmp.security.usm.stats import *
from snmp.requests import *
from snmp.utils import typename
from snmp.v3.message import *
from snmp.v3.requests import *

class RequestError(SNMPException):
    pass

class AuthenticationFailure(RequestError):
    def __init__(self, user):
        errmsg = "The remote engine reports an invalid message signature." \
            f" Check that user \"{user}\" is configured to use the right" \
            " authentication protocol and secret."

        super().__init__(errmsg)

class PrivacyFailure(RequestError):
    def __init__(self, user):
        errmsg = "The remote engine was unable to decrypt this request." \
            f" Check that user \"{user}\" is configured to use the right" \
            " privacy protocol and secret."

        super().__init__(errmsg)

class TimeWindowFailure(RequestError):
    def __init__(self):
        errmsg = "The remote engine has replied with multiple" \
            " NotInTimeWindow reports. It may be necessary to manually" \
            " reconfigure the remote engine with a new engine ID."

        super().__init__(errmsg)

class UnknownUserName(RequestError):
    def __init__(self, user):
        errmsg = f"The remote engine does not recognize user \"{user}\"."
        super().__init__(errmsg)

class UnsupportedSecurityLevel(RequestError):
    def __init__(self, user, securityLevel):
        errmsg = "The remote engine reports that user" \
            f" \"{user}\" does not support {securityLevel}."

        super().__init__(errmsg)

class UnhandledReport(RequestError):
    def __init__(self, varbind):
        self.report = varbind
        errmsg = "The remote engine raised a report that the manager" \
            f" does not know how to handle: {varbind}"
        super().__init__()

class InvalidResponseField(IncomingMessageError):
    def __init__(self, field_name, response_value, request_value):
        errmsg = f"The {field_name} of the response ({response_value})" \
            f" did not match the request ({request_value})."
        super().__init__(errmsg)

class NamespaceMismatch(IncomingMessageError):
    def __init__(self, matched, expected):
        # Use double quotes (default string formatting uses single quotes)
        matched_string = "[\"" + "\", \"".join(matched) + "\"]"

        errmsg = f"The response message was successfully authenticated," \
            f" but not under the expected namespace (\"{expected}\")." \
            f" Matched namespace(s): {matched_string}"
        super().__init__(errmsg)

class SNMPv3MessageTable:
    def __init__(self):
        self.authority = MessageIDAuthority()
        self.listeners = {}

    def hear(self, message, channel):
        messageID = message.header.msgID

        try:
            listener = self.listeners[messageID]
        except KeyError:
            raise IncomingMessageError(f"Unknown message ID: {messageID}")

        listener.hear(message)

    def reserveMessageID(self, listener):
        messageID = self.authority.reserve()
        self.listeners[messageID] = listener
        return messageID

    def releaseMessageID(self, messageID):
        try:
            del self.listeners[messageID]
        except KeyError:
            pass

        self.authority.release(messageID)

class DiscoveryHandler:
    class RefreshTask(SchedulerTask):
        def __init__(self, handler):
            self.handler = handler

        def cancel(self):
            self.handler = None

        def run(self):
            if self.handler is not None:
                self.handler.expireMessage()
                self.handler.sendNewMessage()
                self.handler.scheduleRefresh()

    def __init__(self, handle, callback, scheduler, table, sender, channel):
        self.channel = channel
        self.sender = sender
        self.scheduler = scheduler
        self.table = table

        self.messageID = 0
        self.refreshTask = None
        self.unsent = collections.OrderedDict()

        self.callback = callback
        self.handle = handle

    def hear(self, message):
        self.stopDiscovery()
        self.callback(message.securityEngineID)

    def expireMessage(self):
        self.table.releaseMessageID(self.messageID)

    def sendNewMessage(self):
        self.messageID = self.table.reserveMessageID(self)

        message = SNMPv3Message(
            HeaderData(
                self.messageID,
                self.channel.msgMaxSize,
                MessageFlags(reportable=True),
                SecurityModel.USM,
            ),
            ScopedPDU(self.handle.pdu, b""),
            b"",
            SecurityName(b""),
        )

        self.sender.send(message, self.channel)

    def scheduleRefresh(self, delay=None):
        if delay is None:
            delay = next(iter(self.unsent.values()))

        self.refreshTask = self.RefreshTask(self)
        self.scheduler.schedule(self.refreshTask, delay)

    def dropRequest(self, requestID):
        del self.unsent[requestID]

        if not self.unsent:
            self.stopDiscovery()
            return True
        else:
            return False

    def saveRequest(self, requestID, refreshPeriod):
        self.unsent[requestID] = refreshPeriod

        if self.refreshTask is None:
            self.sendNewMessage()
            self.scheduleRefresh(refreshPeriod)

    def stopDiscovery(self):
        if self.refreshTask is not None:
            self.refreshTask.cancel()
            self.refreshTask = None
            self.expireMessage()

class ExpireTask(SchedulerTask):
    def __init__(self, handle):
        self.handle_ref = weakref.ref(handle)

    def run(self):
        handle = self.handle_ref()

        if handle is not None:
            handle.expire()

class SendTask(SchedulerTask):
    def __init__(self, handle, message, engineID, manager):
        self.cancelled = False
        self.exception = False

        self.message = message
        self.messageID = 0
        self.engineID = engineID

        self.handle_ref = weakref.ref(handle)
        self.manager = manager

    def cancel(self):
        self.cancelled = True

    def raiseException(self):
        self.exception = True

    def run(self):
        if self.messageID != 0:
            self.manager.deallocateMessage(self.messageID)
            self.messageID = 0

        if self.cancelled:
            return None

        handle = self.handle_ref()

        if handle is None or not handle.active():
            return None

        if self.exception:
            handle.expire()
            return None

        self.messageID = self.manager.allocateMessage(
            handle.requestID,
            self.engineID,
        )

        message = self.message \
            .withMessageID(self.messageID) \
            .withEngineID(self.engineID)

        self.manager.sender.send(message, self.manager.channel)
        return self

class RequestState:
    def __init__(self, handle, message, refreshPeriod):
        self.handle_ref = weakref.ref(handle)
        self.message = message
        self.refreshPeriod = refreshPeriod

        self.cancelled = {}
        self.tasks = {}

    @property
    def handle(self):
        handle = self.handle_ref()

        if handle is None:
            raise SNMPLibraryBug("Request State was not dropped"
                " when the request handle deactivated")

        return handle

    def cancel(self, engineID):
        if engineID in self.cancelled:
            return False

        task = self.tasks.pop(engineID)
        task.cancel()
        self.cancelled[engineID] = task
        return True

    def expireOnRefresh(self, engineID):
        self.tasks[engineID].raiseException()

    def resend(self, engineID, manager):
        if engineID in self.tasks:
            self.tasks[engineID].cancel()
            del self.tasks[engineID]

        self.send(engineID, manager)

    def select(self, engineID):
        for taskEngineID, task in self.tasks.items():
            if taskEngineID != engineID:
                task.cancel()

    def send(self, engineID, manager):
        if engineID in self.tasks:
            return

        sendTask = SendTask(self.handle, self.message, engineID, manager)
        manager.scheduler.schedule(sendTask, period=self.refreshPeriod)
        self.tasks[engineID] = sendTask

class SNMPv3Manager:
    def __init__(self,
        scheduler,
        table,
        sender,
        channel,
        namespace,
        defaultUserName,
        defaultSecurityLevel,
        engineID,
        autowait,
    ):
        self.channel = channel
        self.sender = sender
        self.table = table

        self.engineIDAuthenticated = False
        self.engineID = engineID
        self.namespace = namespace

        self.autowait = autowait
        self.defaultUserName = defaultUserName
        self.defaultSecurityLevel = defaultSecurityLevel

        self.scheduler = scheduler
        self.requestIDAuthority = RequestIDAuthority()

        self.discovery = None

        self.mapping = {}
        self.requests = {}

    byteStringRegex = re.compile(r"b\'(.*)\'")
    def decodeUserName(self, userName):
        try:
            return userName.decode()
        except UnicodeDecodeError:
            pass

        representation = repr(userName)
        match = re.match(self.byteStringRegex, representation)

        if match is not None:
            return match.group(1)
        else:
            return representation

    def openRequestNoTimeout(self, pdu, *callbacks):
        requestID = self.requestIDAuthority.reserve()
        handle = SNMPv3RequestHandle(
            self.scheduler,
            pdu.withRequestID(requestID),
            # Callbacks
            self.requestIDAuthority.release,
            *callbacks,
        )

        return handle

    def openRequest(self, pdu, timeout, *callbacks):
        handle = self.openRequestNoTimeout(pdu, *callbacks)
        self.scheduler.schedule(ExpireTask(handle), timeout)
        return handle

    def onRequestClosed(self, requestID):
        if self.engineID is None:
            if (self.discovery is not None
            and self.discovery.dropRequest(requestID)):
                self.discovery = None

        del self.requests[requestID]

    def setEngineID(self, engineID, auth):
        if self.engineIDAuthenticated and not auth:
            return

        if auth and not self.engineIDAuthenticated:
            self.engineIDAuthenticated = True

        if engineID == self.engineID:
            return

        oldEngineID = self.engineID
        self.engineID = engineID

        for request in self.requests.values():
            request.select(engineID)
            request.send(engineID, self)

    def discoveryCallback(self, engineID):
        self.discovery = None
        self.setEngineID(engineID, False)

    def allocateMessage(self, requestID, engineID):
        messageID = self.table.reserveMessageID(self)
        self.mapping[messageID] = requestID, engineID
        return messageID

    def deallocateMessage(self, messageID):
        del self.mapping[messageID]
        self.table.releaseMessageID(messageID)

    def hear(self, message):
        messageID = message.header.msgID

        try:
            requestID, engineID = self.mapping[messageID]
        except KeyError as err:
            errmsg = f"RequestID not found for message {messageID}"
            raise SNMPLibraryBug(errmsg) from err

        try:
            requestState = self.requests[requestID]
        except KeyError as err:
            errmsg = f"Request {requestID} is no longer active"
            raise IncomingMessageError(errmsg) from err

        handle = requestState.handle
        requestMessage = requestState.message

        pdu = message.scopedPDU.pdu
        contextName = message.scopedPDU.contextName
        securityLevel = message.header.flags.securityLevel
        securityName = message.securityName

        if pdu.INTERNAL_CLASS:
            if len(pdu.variableBindings) < 1:
                raise IncomingMessageError("No OIDs in report")

            oid = pdu.variableBindings[0].name
            if oid == usmStatsUnknownEngineIDsInstance:
                requestState.send(message.securityEngineID, self)
            elif message.securityEngineID != engineID:
                errmsg = "Engine ID does not match the request"
                raise IncomingMessageError(errmsg)
            elif oid == usmStatsNotInTimeWindowsInstance:
                if requestMessage.header.flags.authFlag:
                    refreshPeriod = requestState.refreshPeriod
                    cancelled = requestState.cancel(engineID)

                    if cancelled:
                        requestState.resend(engineID, self)
                    else:
                        handle.report(TimeWindowFailure())
                        if message.header.flags.authFlag:
                            handle.expire()
                        else:
                            requestState.expireOnRefresh(engineID)
            elif oid == usmStatsUnsupportedSecLevelsInstance:
                requestSecLevel = requestMessage.header.flags.securityLevel

                if securityLevel < requestSecLevel:
                    userName = requestMessage.securityName.userName
                    user = self.decodeUserName(userName)
                    error = UnsupportedSecurityLevel(user, requestSecLevel)
                    handle.report(error)

                    if message.header.flags.authFlag:
                        handle.expire()
                    else:
                        requestState.expireOnRefresh(engineID)
            elif oid == usmStatsUnknownUserNamesInstance:
                if not message.header.flags.authFlag:
                    userName = requestMessage.securityName.userName
                    user = self.decodeUserName(userName)
                    handle.report(UnknownUserName(user))
                    requestState.expireOnRefresh(engineID)
            elif oid == usmStatsWrongDigestsInstance:
                if requestMessage.header.flags.authFlag:
                    userName = requestMessage.securityName.userName
                    user = self.decodeUserName(userName)
                    handle.report(AuthenticationFailure(user))
                    requestState.expireOnRefresh(engineID)
            elif oid == usmStatsDecryptionErrorsInstance:
                if requestMessage.header.flags.privFlag:
                    userName = requestMessage.securityName.userName
                    user = self.decodeUserName(userName)
                    handle.report(PrivacyFailure(user))

                    if message.header.flags.authFlag:
                        handle.expire()
                    else:
                        requestState.expireOnRefresh(engineID)
            else:
                handle.report(UnhandledReport(pdu.variableBindings[0]))

                if message.header.flags.authFlag:
                    handle.expire()
                else:
                    requestState.expireOnRefresh(engineID)
        else:
            if pdu.requestID != requestID:
                raise InvalidResponseField(
                    "requestID",
                    pdu.requestID,
                    requestID,
                )
            elif securityLevel < requestMessage.header.flags.securityLevel:
                raise InvalidResponseField(
                    "securityLevel",
                    securityLevel,
                    requestMessage.header.flags.securityLevel,
                )
            elif message.scopedPDU.contextEngineID != engineID:
                raise InvalidResponseField(
                    "contextEngineID",
                    message.scopedPDU.contextEngineID,
                    engineID,
                )
            elif contextName != requestMessage.scopedPDU.contextName:
                raise InvalidResponseField(
                    "contextName",
                    contextName,
                    requestMessage.scopedPDU.contextName,
                )
            elif message.securityEngineID != engineID:
                raise InvalidResponseField(
                    "securityEngineID",
                    message.securityEngineID,
                    engineID,
                )
            elif securityName.userName != requestMessage.securityName.userName:
                raise InvalidResponseField(
                    "securityName",
                    securityName.userName,
                    requestMessage.securityName.userName,
                )
            elif (requestMessage.header.flags.authFlag
            and self.namespace not in securityName.namespaces):
                raise NamespaceMismatch(
                    securityName.namespaces,
                    self.namespace,
                )

            self.setEngineID(
                message.securityEngineID,
                message.header.flags.authFlag,
            )

            handle.push(pdu)

    def makeRequest(self,
        pdu,
        user=None,
        securityLevel=None,
        context=b"",
        wait=None,
        timeout=10.0,
        refreshPeriod=1.0,
    ):
        if wait is None:
            wait = self.autowait

        if securityLevel is None:
            securityLevel = self.defaultSecurityLevel

        if user is None:
            userName = self.defaultUserName
        else:
            try:
                userName = user.encode()
            except AttributeError as err:
                errmsg = f"'user' must be a str, not {typename(user)}"
                raise TypeError(errmsg) from err

        handle = self.openRequest(pdu, timeout)

        if handle.active():
            header = HeaderData(
                0,
                self.channel.msgMaxSize,
                MessageFlags(securityLevel, True),
                SecurityModel.USM,
            )

            scopedPDU = ScopedPDU(handle.pdu, b"", context)
            securityName = SecurityName(userName, self.namespace)
            message = SNMPv3Message(header, scopedPDU, b"", securityName)

            requestState = RequestState(handle, message, refreshPeriod)
            self.requests[handle.requestID] = requestState

            if self.engineID is None:
                if self.discovery is None:
                    self.discovery = DiscoveryHandler(
                        self.openRequestNoTimeout(GetRequestPDU()),
                        self.discoveryCallback,
                        self.scheduler,
                        self.table,
                        self.sender,
                        self.channel,
                    )

                self.discovery.saveRequest(handle.requestID, refreshPeriod)
            else:
                requestState.send(self.engineID, self)

            handle.addCallback(self.onRequestClosed)

        if wait:
            return handle.wait()
        else:
            return handle

    def get(self, *oids, **kwargs):
        pdu = GetRequestPDU(*oids)
        return self.makeRequest(pdu, **kwargs)

    def getBulk(self, *oids, nonRepeaters=0, maxRepetitions=1, **kwargs):
        pdu = GetBulkRequestPDU(
            *oids,
            nonRepeaters=nonRepeaters,
            maxRepetitions=maxRepetitions,
        )

        return self.makeRequest(pdu, **kwargs)

    def getNext(self, *oids, **kwargs):
        pdu = GetNextRequestPDU(*oids)
        return self.makeRequest(pdu, **kwargs)

    def set(self, *varbinds, **kwargs):
        pdu = SetRequestPDU(*varbinds)
        return self.makeRequest(pdu, **kwargs)



__all__ = ["SignedUsmParameters", "UnsignedUsmParameters"]

from snmp.ber import *
from snmp.smi import *
from snmp.utils import *

class UnsignedUsmParameters(Sequence):
    def __init__(self,
        engineID,
        engineBoots,
        engineTime,
        userName,
        padding,
        salt,
    ):
        if engineBoots < 0:
            raise ValueError(f"negative value for engineBoots: {engineBoots}")
        elif engineTime < 0:
            raise ValueError(f"negative value for engineTime: {engineTime}")
        elif len(userName) > 32:
            raise ValueError(f"userName exceeds 32 characters: {userName!r}")

        self._engineID = OctetString(engineID)
        self._engineBoots = Integer(engineBoots)
        self._engineTime = Integer(engineTime)
        self._userName = OctetString(userName)
        self._padding = OctetString(padding)
        self._salt = OctetString(salt)

    @property
    def engineID(self):
        return self._engineID.data

    @property
    def engineBoots(self):
        return self._engineBoots.value

    @property
    def engineTime(self):
        return self._engineTime.value

    @property
    def userName(self):
        return self._userName.data

    @property
    def padding(self):
        return self._padding.data

    @property
    def salt(self):
        return self._salt.data

    def __iter__(self):
        yield self._engineID
        yield self._engineBoots
        yield self._engineTime
        yield self._userName
        yield self._padding
        yield self._salt

    def __len__(self):
        return 6

    def __repr__(self):
        args = (
            str(self.engineID),
            str(self.engineBoots),
            str(self.engineTime),
            str(self.userName),
            str(self.padding),
            str(self.salt),
        )

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self):
        return self.toString()

    def toString(self, depth = 0, tab = "    "):
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Authoritative Engine ID: {self.engineID!r}",
            f"{subindent}Authoritative Engine Boots: {self.engineBoots}",
            f"{subindent}Authoritative Engine Time: {self.engineTime}",
            f"{subindent}User Name: {self.userName!r}",
            f"{subindent}Authentication Padding: {self.padding!r}",
            f"{subindent}Encryption Salt: {self.salt!r}",
        ))

    @classmethod
    def deserialize(cls, data):
        engineID, ebdata = OctetString.decode(data)
        engineBoots, etdata = Integer.decode(ebdata)

        if engineBoots.value < 0:
            errmsg = f"negative value for engineBoots: {engineBoots.value}"
            raise ParseError(errmsg, ebdata, etdata)

        engineTime, undata = Integer.decode(etdata)

        if engineTime.value < 0:
            errmsg = f"negative value for engineTime: {engineTime.value}"
            raise ParseError(errmsg, etdata, undata)

        userName, ptr = OctetString.decode(undata)

        if len(userName.data) > 32:
            errmsg = f"userName exceeds 32 characters: {userName.data!r}"
            raise ParseError(errmsg, undata, ptr)

        padding, ptr    = OctetString.decode(ptr)
        salt            = OctetString.decodeExact(ptr)

        try:
            return cls(
                engineID.data,
                engineBoots.value,
                engineTime.value,
                userName.data,
                padding.data,
                salt.data,
            )
        except ValueError as err:
            raise cls.DeserializeError(*err.args) from err

    @classmethod
    def findPadding(self, msgSecurityParameters):
        tag, ptr, tail          = decode(msgSecurityParameters)
        tag, engineID, ptr      = decode(ptr)
        tag, engineBoots, ptr   = decode(ptr)
        tag, engineTime, ptr    = decode(ptr)
        tag, userName, ptr      = decode(ptr)
        tag, ptr, tail          = decode(ptr)
        return ptr

class SignedUsmParameters(Sequence):
    def __init__(self,
        engineID,
        engineBoots,
        engineTime,
        userName,
        signature,
        salt,
    ):
        if engineBoots < 0:
            raise ValueError(f"negative value for engineBoots: {engineBoots}")
        elif engineTime < 0:
            raise ValueError(f"negative value for engineTime: {engineTime}")
        elif len(userName) > 32:
            raise ValueError(f"userName exceeds 32 characters: {userName!r}")

        self._engineID = OctetString(engineID)
        self._engineBoots = Integer(engineBoots)
        self._engineTime = Integer(engineTime)
        self._userName = OctetString(userName)
        self._signature = OctetString(signature)
        self._salt = OctetString(salt)

    @property
    def engineID(self):
        return self._engineID.data

    @property
    def engineBoots(self):
        return self._engineBoots.value

    @property
    def engineTime(self):
        return self._engineTime.value

    @property
    def userName(self):
        return self._userName.data

    @property
    def signature(self):
        return self._signature.original

    @property
    def salt(self):
        return self._salt.data

    def __iter__(self):
        yield self._engineID
        yield self._engineBoots
        yield self._engineTime
        yield self._userName
        yield self._signature
        yield self._salt

    def __len__(self):
        return 6

    def __repr__(self):
        args = (
            str(self.engineID),
            str(self.engineBoots),
            str(self.engineTime),
            str(self.userName),
            repr(self.signature),
            str(self.salt),
        )

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self):
        return self.toString()

    def toString(self, depth = 0, tab = "    "):
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Authoritative Engine ID: {self.engineID!r}",
            f"{subindent}Authoritative Engine Boots: {self.engineBoots}",
            f"{subindent}Authoritative Engine Time: {self.engineTime}",
            f"{subindent}User Name: {self.userName!r}",
            f"{subindent}Signature: {self.signature[:]!r}",
            f"{subindent}Encryption Salt: {self.salt!r}",
        ))

    @classmethod
    def deserialize(cls, data):
        engineID, ebdata   = OctetString.decode(data)
        engineBoots, etdata = Integer.decode(ebdata)

        if engineBoots.value < 0:
            errmsg = f"negative value for engineBoots: {engineBoots.value}"
            raise ParseError(errmsg, ebdata, etdata)

        engineTime, undata = Integer.decode(etdata)

        if engineTime.value < 0:
            errmsg = f"negative value for engineTime: {engineTime.value}"
            raise ParseError(errmsg, etdata, undata)

        userName, ptr = OctetString.decode(undata)

        if len(userName.data) > 32:
            errmsg = f"userName exceeds 32 characters: {userName.data!r}"
            raise ParseError(errmsg, undata, ptr)

        signature, ptr  = OctetString.decode(ptr, copy=False)
        salt            = OctetString.decodeExact(ptr)

        try:
            return cls(
                engineID.data,
                engineBoots.value,
                engineTime.value,
                userName.data,
                signature.original,
                salt.data,
            )
        except ValueError as err:
            raise cls.DeserializeError(*err.args) from err



__all__ = [
    "Integer", "Integer32", "Unsigned", "Unsigned32",
    "Counter32", "Gauge32", "TimeTicks", "Counter64",
    "OctetString", "IpAddress", "Opaque",
    "NULL", "Null", "OID", "zeroDotZero",
    "Sequence", "VarBind", "VarBindList",
    "NoSuchObject", "NoSuchInstance", "EndOfMibView",
]

from socket import inet_aton, inet_ntoa

from snmp.asn1 import *
from snmp.ber import *
from snmp.utils import *

class BoundedInteger(INTEGER):
    def __init__(self, value):
        if not self.inRange(value):
            raise ValueError(f"{value} is out of range for {typename(self)}")

        super().__init__(value)

    @classmethod
    def construct(cls, value):
        try:
            return cls(value)
        except ValueError as err:
            raise cls.DeserializeError(err.args[0])

    @classmethod
    def deserialize(cls, data):
        value = int.from_bytes(data, cls.BYTEORDER, signed=cls.SIGNED)
        return cls.construct(value)

    @classmethod
    def inRange(cls, value):
        if value < 0 and not cls.SIGNED:
            return False

        nbits = cls.bitCount(value)
        allowable = cls.BITS - 1 if cls.SIGNED else cls.BITS
        return nbits <= allowable

class Integer32(BoundedInteger):
    BITS = 32
    SIGNED  = True

class Unsigned32(BoundedInteger):
    BITS = 32
    SIGNED = False
    TAG = Tag(2, cls = Tag.Class.APPLICATION)

Integer = Integer32
Unsigned = Unsigned32

class Counter32(Unsigned32):
    TAG = Tag(1, cls = Tag.Class.APPLICATION)

class Gauge32(Unsigned32):
    pass

class TimeTicks(Unsigned32):
    TAG = Tag(3, cls = Tag.Class.APPLICATION)

class Counter64(BoundedInteger):
    BITS = 64
    SIGNED = False
    TAG = Tag(6, cls = Tag.Class.APPLICATION)

class OctetString(OCTET_STRING):
    def __init__(self, data = b""):
        self.check(data)

        self._original = data
        super().__init__(data[:])

    def __repr__(self):
        return f"{typename(self)}({self.original!r})"

    @property
    def original(self):
        return self._original

    @classmethod
    def check(cls, data):
        if len(data) > 0xffff:
            raise ValueError(f"{typename(cls)} is limited to {0xffff} bytes")

    @classmethod
    def construct(cls, data):
        try:
            return cls(data)
        except ValueError as err:
            raise cls.DeserializeError(err.args[0]) from err

class IpAddress(OCTET_STRING):
    TAG = Tag(0, cls = Tag.Class.APPLICATION)

    def __init__(self, addr):
        try:
            data = inet_aton(addr)
        except OSError as err:
            raise ValueError(f"Invalid IPv4 address: \"{addr}\"") from err

        super().__init__(data)
        self.addr = addr

    def __repr__(self):
        return f"{typename(self)}({repr(self.addr)})"

    def asOID(self, implied = False):
        return super().asOID(implied=True)

    @classmethod
    def fromBytes(cls, data):
        try:
            addr = inet_ntoa(data)
        except OSError as err:
            errmsg = f"Invalid IPv4 address encoding: {data!r}"
            raise ValueError(errmsg) from err

        return cls(addr)

    @classmethod
    def fromOID(cls, nums, implied = False):
        data = bytes(next(nums) for i in range(4))
        return cls.fromBytes(data)

    @classmethod
    def construct(cls, data):
        try:
            return cls.fromBytes(data[:])
        except ValueError as err:
            raise cls.DeserializeError(err.args[0]) from err

class Opaque(OCTET_STRING):
    TAG = Tag(4, cls = Tag.Class.APPLICATION)

Null = NULL

class NoSuchObject(Null):
    TAG = Tag(0, cls = Tag.Class.CONTEXT_SPECIFIC)

class NoSuchInstance(Null):
    TAG = Tag(1, cls = Tag.Class.CONTEXT_SPECIFIC)

class EndOfMibView(Null):
    TAG = Tag(2, cls = Tag.Class.CONTEXT_SPECIFIC)

class OID(OBJECT_IDENTIFIER):
    def __init__(self, *subidentifiers):
        if len(subidentifiers) > 128:
            errmsg = "OID may not contain more than 128 sub-identifiers"
            raise ValueError(errmsg)

        if any(map(lambda x: x.bit_length() > 32, subidentifiers)):
            raise ValueError("Sub-identifiers are limited to 32-bits unsigned")

        super().__init__(*subidentifiers)

    def getIndex(self, prefix, cls = Integer, implied = False):
        return self.decodeIndex(prefix, cls, implied=implied)[0]

zeroDotZero = OID(0, 0)

class Sequence(SEQUENCE):
    pass

class VarBind(Sequence):
    TYPES = {
        cls.TAG: cls for cls in (
            Integer,
            OctetString,
            Null,
            OID,
            IpAddress,
            Counter32,
            Gauge32,
            TimeTicks,
            Opaque,
            Counter64,
            NoSuchObject,
            NoSuchInstance,
            EndOfMibView,
        )
    }

    def __init__(self, name, value = None):
        if not isinstance(name, OID):
            name = OID.parse(name)

        if value is None:
            value = Null()

        self.name = name
        self.value = value

    def __iter__(self):
        yield self.name
        yield self.value

    def __len__(self):
        return 2

    def __repr__(self):
        return f"{typename(self)}({self.name!r}, {self.value!r})"

    def __str__(self):
        return f"{self.name}: {self.value}"

    @classmethod
    def deserialize(cls, data):
        name, data = OID.decode(data)
        tag, _ = Tag.decode(data)

        try:
            valueType = cls.TYPES[tag]
        except KeyError as err:
            errmsg = f"Invalid variable value type: {tag}"
            raise ParseError(errmsg, data) from err

        return cls(name, valueType.decodeExact(data))

class VarBindList(Sequence):
    def __init__(self, *args, unpack=False):
        self.variables = tuple(
            vb if isinstance(vb, VarBind) else
            VarBind(*vb) if unpack else VarBind(vb)
            for vb in args
        )

    def __bool__(self):
        return bool(self.variables)

    def __getitem__(self, key):
        return self.variables[key]

    def __iter__(self):
        return iter(self.variables)

    def __len__(self):
        return len(self.variables)

    def __repr__(self):
        args = ", ".join(repr(var) for var in self.variables)
        return f"{typename(self)}({args})"

    def __str__(self):
        return self.toString()

    def toString(self, indent = ""):
        return "\n".join(f"{indent}{var}" for var in self.variables)

    @classmethod
    def deserialize(cls, data):
        objects = []

        while data:
            var, data = VarBind.decode(data)
            objects.append(var)

        return cls(*objects)



__all__ = ["UdpMultiplexor"]

import select
import socket

from snmp.transport import *

STOPMSG = bytes(1)

class GenericUdpMultiplexor(TransportMultiplexor):
    def __init__(self):
        self.sockets = {}

        domain = TransportDomain.UDP_IPv4
        self.r = socket.socket(domain.address_family, socket.SOCK_DGRAM)
        self.w = socket.socket(domain.address_family, socket.SOCK_DGRAM)

        self.r.setblocking(False)
        self.r.bind((domain.loopback_address, 0))
        self.w.bind((domain.loopback_address, 0))

        self.readfds = [self.r.fileno()]

    def register(self, sock, listener):
        self.readfds.append(sock.fileno)
        self.sockets[sock.fileno] = sock, listener

    def poll(self, timeout = None):
        interrupted = False

        if self.readfds:
            ready = select.select(self.readfds, [], [], timeout)[0]
            for fd in ready:
                try:
                    sock, listener = self.sockets[fd]
                except KeyError:
                    if fd == self.r.fileno():
                        data, addr = self.r.recvfrom(len(STOPMSG))
                        if addr == self.w.getsockname() and data == STOPMSG:
                            interrupted = True
                else:
                    addr, data = sock.receive()
                    listener.hear(data, TransportChannel(sock, addr))

        return interrupted

    def stop(self):
        if self.w is not None:
            self.w.sendto(STOPMSG, self.r.getsockname())

    def close(self):
        for sock, _ in self.sockets.values():
            sock.close()

        if self.w is not None:
            self.w.close()
            self.r.close()

UdpMultiplexor = GenericUdpMultiplexor



__all__ = ["EVP_CIPHER", "EVP_CIPHER_CTX", "ffi", "lib"]

from snmp.cffi import *
from snmp.typing import *

class EVP_CIPHER:
    ...

class EVP_CIPHER_CTX:
    ...

class OpenSSL_FFI(FFI):
    NULL: Pointer[Any]

    @overload
    def new(self, typename: Literal["int*"]) -> Pointer[int]:
        ...

    @overload
    def new(self,
        typename: Literal["unsigned char[]"],
        n: int,
    ) -> UnsignedCharArray:
        ...

class OpenSSL_LIB:
    def EVP_CIPHER_CTX_new(self) -> Pointer[EVP_CIPHER_CTX]:
        ...

    def EVP_CIPHER_CTX_free(self, ctx: Pointer[EVP_CIPHER_CTX]) -> None:
        ...

    def EVP_DecryptInit(self,
        ctx: Pointer[EVP_CIPHER_CTX],
        type: Pointer[EVP_CIPHER],
        key: bytes,
        iv: bytes,
    ) -> int:
        ...

    def EVP_DecryptUpdate(self,
        ctx: Pointer[EVP_CIPHER_CTX],
        out: UnsignedCharArray,
        outl: Pointer[int],
        _in: bytes,
        inl: int,
    ) -> int:
        ...

    def EVP_EncryptInit(self,
        ctx: Pointer[EVP_CIPHER_CTX],
        type: Pointer[EVP_CIPHER],
        key: bytes,
        iv: bytes,
    ) -> int:
        ...

    def EVP_EncryptUpdate(self,
        ctx: Pointer[EVP_CIPHER_CTX],
        out: UnsignedCharArray,
        outl: Pointer[int],
        _in: bytes,
        inl: int,
    ) -> int:
        ...

    def EVP_aes_128_cfb128(self) -> Pointer[EVP_CIPHER]:
        ...

    def EVP_des_cbc(self) -> Pointer[EVP_CIPHER]:
        ...

ffi = OpenSSL_FFI()
lib = OpenSSL_LIB()



__all__ = ["SNMPv3Interpreter", "SNMPv3MessageSorter"]

import weakref

from snmp.smi import OID
from snmp.pdu import ReportPDU
from snmp.security import SecurityModel

from .message import *

snmpUnknownPDUHandlersInstance = OID.parse("1.3.6.1.6.3.11.2.1.3")

class SNMPv3Interpreter:
    def __init__(self, usm):
        self.usm = usm

    def decode(self, data):
        message = SNMPv3WireMessage.decodeExact(data)
        return self.usm.processIncoming(message)

    def encode(self, message):
        return self.usm.prepareOutgoing(message)

    def makeReport(self, message, *varbinds):
        return SNMPv3Message(
            HeaderData(
                message.header.msgID,
                message.header.maxSize,
                MessageFlags(message.header.flags.securityLevel),
                SecurityModel.USM,
            ),
            ScopedPDU(
                ReportPDU(
                    *varbinds,
                    requestID=message.scopedPDU.pdu.requestID,
                ),
                message.scopedPDU.contextEngineID,
                message.scopedPDU.contextName,
            ),
            message.securityEngineID,
            message.securityName,
        )

    def pduType(self, message):
        return type(message.scopedPDU.pdu)

class SNMPv3MessageSorter:
    def __init__(self, interpreter):
        self.interpreter = interpreter
        self.subscribers = weakref.WeakValueDictionary()

        self.unknownHandlers = 0

    def register(self, pduType, subscriber):
        subscribed = self.subscribers.setdefault(pduType.TAG, subscriber)
        return subscriber is subscribed

    def forward(self, message, channel):
        pduType = self.interpreter.pduType(message)

        try:
            subscriber = self.subscribers[pduType.TAG]
        except KeyError:
            if not pduType.RESPONSE_CLASS:
                self.unknownHandlers += 1
                if pduType.CONFIRMED_CLASS:
                    try:
                        reportMessage = self.interpreter.makeReport(
                            message,
                            VarBind(
                                snmpUnknownPDUHandlersInstance,
                                Counter32(self.unknownHandlers),
                            )
                        )

                        data = self.interpreter.encode(reportMessage)
                        channel.send(data)
                    except Exception:
                        pass
        else:
            subscriber.hear(message, channel)

    def hear(self, data, channel):
        try:
            message = self.interpreter.decode(data)
        except ReportMessage as report:
            self.send(report.message, channel)
        else:
            self.forward(message, channel)

    def send(self, message, channel):
        channel.send(self.interpreter.encode(message))



__all__ = ["Catcher", "VersionDecoder"]

import logging
import os
import re
import weakref

from snmp.exception import *
from snmp.message import *
from snmp.utils import *

class Catcher:
    def __init__(self, listener, verbose=False):
        self.listener = listener
        self.logger = logging.getLogger(__name__.split(".")[0])
        self.verbose = verbose

        self.packets = 0

    def hear(self, data, channel):
        self.packets += 1

        try:
            self.listener.hear(data, channel)
        except IncomingMessageErrorWithPointer as err:
            if self.verbose:
                self.logger.debug(f"{typename(err)}:{err}")
        except IncomingMessageError as err:
            if self.verbose:
                hexdump = re.sub(r"(?<=.{2})(.{2})", r" \1", data.hex())
                self.logger.debug(f"{err!r}{os.linesep}{hexdump}")
        except Exception as exc:
            self.logger.exception(exc)

class VersionDecoder:
    def __init__(self):
        self.listeners = weakref.WeakValueDictionary()

    def hear(self, data, channel):
        msgVersion = VersionOnlyMessage.decodeExact(data).version

        try:
            listener = self.listeners[msgVersion]
        except KeyError as err:
            errmsg = f"Ignoring {msgVersion.name} message" \
                " because no application is listening for it."
            errdata = subbytes(data)
            raise BadVersion(errmsg, errdata) from err
        else:
            listener.hear(data, channel)

    def register(self, version, listener):
        registered = self.listeners.setdefault(version, listener)
        return registered is listener



__all__ = ["DesCbc"]

import os

from Crypto.Cipher import DES

from snmp.smi import OID
from snmp.security.usm import PrivProtocol

class DesCbc(PrivProtocol):
    BYTEORDER = "big"

    BLOCKLEN = 8
    KEYLEN = BLOCKLEN * 2
    SALTLEN = BLOCKLEN // 2
    SALTWRAP = 1 << (8 * SALTLEN)

    def __init__(self, key):
        if len(key) < self.KEYLEN:
            raise ValueError(f"key must be at least {self.KEYLEN} bytes long")

        self.algorithm = OID.parse("1.3.6.1.6.3.10.1.2.2")
        self.key = key[:self.BLOCKLEN]
        self.preIV = key[self.BLOCKLEN:self.KEYLEN]
        self.salt = int.from_bytes(os.urandom(self.SALTLEN), self.BYTEORDER)

    def __eq__(self, other):
        try:
            return (self.algorithm == other.algorithm
                and self.key == other.key
                and self.preIV == other.preIV)
        except AttributeError:
            return NotImplemented

    def computeIV(self, salt):
        return bytes(a ^ b for a, b in zip(self.preIV, salt))

    def newCipher(self, iv):
        return DES.new(self.key, DES.MODE_CBC, iv=iv)

    def pad(self, data):
        n = self.BLOCKLEN - (len(data) % self.BLOCKLEN)
        return data + bytes(n)

    def decrypt(self, data, engineBoots, engineTime, salt):
        if len(data) % self.BLOCKLEN:
            errmsg = "DES ciphertext must be a multiple of {} in length"
            raise ValueError(errmsg.format(self.BLOCKLEN))

        return self.newCipher(self.computeIV(salt)).decrypt(self.pad(data))

    def encrypt(self, data, engineBoots, engineTime):
        self.salt = (self.salt + 1) % self.SALTWRAP
        salt = b''.join((
            engineBoots.to_bytes(self.BLOCKLEN - self.SALTLEN, self.BYTEORDER),
            self.salt.to_bytes(self.SALTLEN, self.BYTEORDER),
        ))

        iv = self.computeIV(salt)
        ciphertext = self.newCipher(iv).encrypt(self.pad(data))

        return ciphertext, salt



__all__ = [
    "BadVersion", "InvalidMessage",
    "Message", "ProtocolVersion", "VersionOnlyMessage",
]

from snmp.exception import IncomingMessageErrorWithPointer

from .core import *
from .version import *

class InvalidMessage(IncomingMessageErrorWithPointer):
    pass



__all__ = ["subbytes", "typename"]

import os

class subbytes:
    """Operate on a slice of a bytes-like object without copying any data.

    This class represents a sequence of bytes, similarly to how data is
    represented in a bytes-like object. Unlike a bytes-like object, however,
    a subbytes object does not store this data in a memory block of its own.
    It holds a reference to a real bytes-like object, and acts on a
    sub-sequence of that object's data, as if it had been copied into a
    separate bytes-like object. The arguments to the constructor mirror the
    arguments to the built-in slice() function, (without the step parameter)
    so that subbytes(data, start, stop) represents the same sequence of bytes
    as data[start:stop].

    .. attribute:: data

        The data attribute provides a reference to the underlying bytes-like
        object. If the data argument to the constructor is an instance of
        subbytes, it will be unwrapped so that this attribute always references
        a bytes-like object directly.
    """

    def __init__(self, data, start = None, stop = None):
        self._data: bytes
        if isinstance(data, subbytes):
            self._data = data.data
            base = data
        else:
            self._data = data
            self._start = 0
            self._stop = len(data)
            base = self

        newstart = base.start if start is None else base.translate(start, True)
        newstop  = base.stop  if stop  is None else base.translate(stop,  True)

        self._start = newstart
        self._stop  = newstop

    @property
    def data(self):
        return self._data

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    def __bool__(self):
        """Indicate that the sequence is non-empty."""
        return self.stop > self.start

    def __eq__(self, other):
        """Compare two sequences of bytes for equality."""
        try:
            # TypeError if other is not Sized
            if len(self) != len(other):
                return False

            # TypeError if other is not Iterable
            for left, right in zip(self, other):
                if left != right:
                    return False
        except TypeError:
            return NotImplemented

        return True

    def __getitem__(self, key):
        """Retrieve an individual byte or copy a sub-sequence."""
        if isinstance(key, int):
            key = self.translate(key)
        elif isinstance(key, slice):
            start, stop = key.start, key.stop

            key = slice(
                self.start if start is None else self.translate(start, True),
                self.stop  if stop  is None else self.translate(stop, True),
                key.step
            )

        return self.data[key]

    def __iter__(self):
        """Produce an iterator for this sequence."""
        for index in range(self.start, self.stop):
            yield self.data[index]

    def __len__(self):
        """Query the length of the sequence."""
        return self.stop - self.start

    @staticmethod
    def _format(data, step, indent="", sep=""):
        lines = []
        start = 0

        while start < len(data):
            stop = start + step
            block = data[start:stop]
            line = sep.join(f"{b:02x}" for b in block)
            lines.append(indent + line)
            start = stop

        return os.linesep.join(lines)

    def __str__(self):
        sections = []

        if self.start > 0:
            prefix = self.data[:self.start]
            sections.append(self._format(prefix, 40))

        indent = 4 * " "
        if self.start < self.stop:
            data = self.data[self.start:self.stop]
            sections.append(self._format(data, 24, indent=indent, sep=" "))
        else:
            sections.append(indent + "(empty)")

        if self.stop < len(self.data):
            suffix = self.data[self.stop:]
            sections.append(self._format(suffix, 40))

        separator = 2 * os.linesep
        body = separator.join(sections)

        header = 32 * "=" + " begin subbytes " + 32 * "="
        footer = 33 * "=" +  " end subbytes "  + 33 * "="
        return os.linesep.join((header, body, footer))

    def __repr__(self):
        """Provide an eval()-able representation of this object."""
        args = [repr(self.data)]

        if self.start:
            args.append(f"start={self.start}")

        if self.stop < len(self.data):
            args.append(f"stop={self.stop}")

        return f"{typename(self)}({', '.join(args)})"

    def translate(self, index, clamp = False):
        """Convert an index of self into an index of self.data.

        This method accepts an integral index of any value and translates it
        into a meaningful index of self.data. If -len(self) <= index <
        len(self), then the result is guaranteed to be a valid index (meaning
        it will not cause an IndexError). If index < -len(self) and clamp is
        True, then this method will return the index of the start of this
        sequence, which is only guaranteed to be valid if the sequence is
        non-empty, and may or may not be valid if the sequence is empty. If
        len(self) <= index and clamp is True, then this method will return the
        stop index, which is the index just beyond the end of the sequence. The
        stop index may or may not be valid, regardless of the length of the
        sequence, but it is guaranteed to match the start index if the sequence
        is empty. Under any other scenario (i.e. index is out of range and
        clamp is False), the result is guaranteed to be an invalid index.

        :meta private:
        """
        if index < 0:
            index += self.stop
            if index < self.start:
                if clamp:
                    return self.start
                else:
                    return len(self.data)
            else:
                return index
        else:
            index += self.start
            if index >= self.stop:
                if clamp:
                    return self.stop
                else:
                    return len(self.data)
            else:
                return index

    def dereference(self):
        """Retrieve the first byte of the sequence.

        If the sequence is empty, an IndexError will be raised.
        """
        return self[0]

    def pop_front(self):
        """Shortcut for split(1) followed by dereference().

        This method returns a tuple containing, first, the int value of the
        first byte, and second, a new subbytes object referencing the same
        underlying sequence, starting from index 1.

        If the sequence is empty, this method will raise an IndexError.
        """
        a, b = self.split(1)
        return a.dereference(), b

    def replace(self, replacement):
        """Substitute the given string in place of the current sequence

        This method produces a copy of the wrapped object, in which the
        current sequence has been replaced with the contents of the
        replacement argument.
        """
        return self.data[:self.start] + replacement + self.data[self.stop:]

    def split(self, index):
        """Split the sequence at the given index.

        Return two new objects, the first referencing the portion of the
        sequence  starting at the beginning of the current sequence and ending
        just before index, and the second referencing the portion beginning
        at index and ending at the end of the current sequence.
        """
        return subbytes(self, stop=index), subbytes(self, start=index)

def typename(cls, qualified = False):
    """Query an object to determine its type.

    :param cls:
        If cls is a class, this function will return its name. If cls is an
        object, this function will return the name of the object's class.
    :param bool qualified:
        Indicate whether to return the fully-qualified name, which includes
        the module path, as well as the names of any enclosing classes (for
        inner classes).
    """
    if not isinstance(cls, type):
        cls = type(cls)

    if qualified:
        return ".".join((cls.__module__, cls.__qualname__))
    else:
        return cls.__name__



__all__ = [
    "usmStatsUnsupportedSecLevelsInstance",
    "usmStatsNotInTimeWindowsInstance",
    "usmStatsUnknownUserNamesInstance",
    "usmStatsUnknownEngineIDsInstance",
    "usmStatsWrongDigestsInstance",
    "usmStatsDecryptionErrorsInstance",
]

from snmp.smi import OID

usmStats = OID.parse("1.3.6.1.6.3.15.1.1")
usmStatsUnsupportedSecLevelsInstance= usmStats.extend(1, 0)
usmStatsNotInTimeWindowsInstance    = usmStats.extend(2, 0)
usmStatsUnknownUserNamesInstance    = usmStats.extend(3, 0)
usmStatsUnknownEngineIDsInstance    = usmStats.extend(4, 0)
usmStatsWrongDigestsInstance        = usmStats.extend(5, 0)
usmStatsDecryptionErrorsInstance    = usmStats.extend(6, 0)



__all__ = ["EngineTime", "TimeKeeper", "UsmNotInTimeWindow"]

from snmp.exception import IncomingMessageError

class UsmNotInTimeWindow(IncomingMessageError):
    pass

class EngineTime:
    class NegativeEngineBoots(ValueError):
        pass

    class PriorToReboot(UsmNotInTimeWindow):
        pass

    class ReconfigRequired(UsmNotInTimeWindow):
        pass

    class TooEarly(UsmNotInTimeWindow):
        pass

    class TooOld(UsmNotInTimeWindow):
        pass

    MAX_ENGINE_BOOTS = (1 << 31) - 1
    TIME_WINDOW_SIZE = 150

    def __init__(self, timestamp, snmpEngineBoots = 0, authoritative = False):
        self.authenticated = False
        self.authoritative = authoritative
        self.latestBootTime = timestamp
        self.snmpEngineBoots = snmpEngineBoots

    @property
    def snmpEngineBoots(self):
        return self._snmpEngineBoots

    @snmpEngineBoots.setter
    def snmpEngineBoots(self, engineBoots):
        if engineBoots < 0:
            raise self.NegativeEngineBoots(engineBoots)

        self._snmpEngineBoots = min(engineBoots, self.MAX_ENGINE_BOOTS)

    @property
    def valid(self):
        return self.snmpEngineBoots < self.MAX_ENGINE_BOOTS

    def snmpEngineTime(self, timestamp):
        return int(timestamp - self.latestBootTime)

    def setEngineTime(self, timestamp, engineTime):
        self.latestBootTime = timestamp - engineTime

    def computeAge(self, timestamp, engineTime):
        return self.snmpEngineTime(timestamp) - engineTime

    def hint(self, timestamp, engineBoots, engineTime):
        if not self.authenticated:
            self.snmpEngineBoots = engineBoots
            self.setEngineTime(timestamp, engineTime)

    def update(self, timestamp, engineBoots, engineTime):
        if self.authenticated:
            if engineBoots == self.snmpEngineBoots:
                if engineTime > self.snmpEngineTime(timestamp):
                    self.setEngineTime(timestamp, engineTime)
            elif engineBoots > self.snmpEngineBoots:
                self.snmpEngineBoots = engineBoots
                self.setEngineTime(timestamp, engineTime)
        else:
            self.snmpEngineBoots = engineBoots
            self.setEngineTime(timestamp, engineTime)
            self.authenticated = True

    def verifyTimeliness(self, timestamp, msgBoots, msgTime):
        if msgBoots != self.snmpEngineBoots:
            raise self.PriorToReboot()

        age = self.computeAge(timestamp, msgTime)
        if age > self.TIME_WINDOW_SIZE:
            raise self.TooOld()
        elif self.authoritative and (-1 * age) > self.TIME_WINDOW_SIZE:
            raise self.TooEarly()
        elif not self.valid:
            raise self.ReconfigRequired()

class TimeKeeper:
    def __init__(self):
        self.table = {}

    def assertEntry(self, engineID, timestamp):
        try:
            et = self.table[engineID]
        except KeyError as err:
            et = EngineTime(timestamp)
            self.table[engineID] = et

        return et

    def getEngineTime(self, engineID, timestamp):
        et = self.assertEntry(engineID, timestamp)
        return et.snmpEngineBoots, et.snmpEngineTime(timestamp)

    def hint(self, engineID, timestamp, msgBoots, msgTime):
        et = self.assertEntry(engineID, timestamp)
        et.hint(timestamp, msgBoots, msgTime)

    def updateAndVerify(self, engineID, timestamp, msgBoots, msgTime):
        et = self.assertEntry(engineID, timestamp)
        et.update(timestamp, msgBoots, msgTime)
        et.verifyTimeliness(timestamp, msgBoots, msgTime)



__all__ = ["UdpMultiplexor"]

import math
import os
import select

from snmp.transport import *

class PosixUdpMultiplexor(TransportMultiplexor):
    def __init__(self):
        self.r, self.w = os.pipe()
        self.sockets = {}

        self.poller = select.poll()
        self.poller.register(self.r, select.POLLIN)

    def register(self, sock, listener):
        self.sockets[sock.fileno] = sock, listener
        self.poller.register(sock.fileno, select.POLLIN)

    def poll(self, timeout = None):
        msecs = math.ceil(timeout * 1000) if timeout is not None else None

        interrupted = False
        ready = self.poller.poll(msecs)

        for fd, _ in ready:
            try:
                sock, listener = self.sockets[fd]
            except KeyError:
                if fd == self.r:
                    os.read(fd, 1)
                    interrupted = True
            else:
                addr, data = sock.receive()
                listener.hear(data, TransportChannel(sock, addr))

        return interrupted

    def stop(self):
        os.write(self.w, bytes(1))

    def close(self):
        for sock, _ in self.sockets.values():
            sock.close()

        os.close(self.w)
        os.close(self.r)

UdpMultiplexor = PosixUdpMultiplexor



__all__ = [
    "GetRequestPDU", "GetNextRequestPDU", "GetBulkRequestPDU",
    "SetRequestPDU",
    "ResponsePDU", "ReportPDU",
    "InformRequestPDU", "SNMPv2TrapPDU",
    "ErrorStatus", "ErrorResponse", "NoSuchName", "ImproperResponse",
]

import enum
import os

from snmp.asn1 import Constructed
from snmp.ber import *
from snmp.exception import *
from snmp.smi import *
from snmp.utils import typename

class ErrorStatus(enum.IntEnum):
    noError             = 0
    tooBig              = 1
    noSuchName          = 2
    badValue            = 3
    readOnly            = 4
    genErr              = 5
    noAccess            = 6
    wrongType           = 7
    wrongLength         = 8
    wrongEncoding       = 9
    wrongValue          = 10
    noCreation          = 11
    inconsistentValue   = 12
    resourceUnavailable = 13
    commitFailed        = 14
    undoFailed          = 15
    authorizationError  = 16
    notWritable         = 17
    inconsistentName    = 18

class PDU(Constructed):
    READ_CLASS = False
    WRITE_CLASS = False
    RESPONSE_CLASS = False
    NOTIFICATION_CLASS = False
    INTERNAL_CLASS = False
    CONFIRMED_CLASS = False

    def __init__(self,
        *args,
        requestID = 0,
        errorStatus = ErrorStatus.noError,
        errorIndex = 0,
        variableBindings = None,
    ):
        self.requestID = requestID
        self.errorStatus = errorStatus
        self.errorIndex = errorIndex

        if variableBindings is None:
            self.variableBindings = VarBindList(*args)
        else:
            self.variableBindings = variableBindings

    def __iter__(self):
        yield Integer(self.requestID)
        yield Integer(self.errorStatus)
        yield Integer(self.errorIndex)
        yield self.variableBindings

    def __len__(self):
        return 4

    def __repr__(self):
        arguments = []
        if self.requestID:
            arguments.append("requestID={}".format(self.requestID))

        if self.errorStatus:
            arguments.append("errorStatus={}".format(self.errorStatus))

        if self.errorIndex:
            arguments.append("errorIndex={}".format(self.errorIndex))

        arguments.append(f"variableBindings={self.variableBindings!r}")

        args = ", ".join(arguments)
        return f"{typename(self)}({args})"

    def __str__(self):
        return self.toString()

    def checkResponse(self, response):
        if not self.validResponse(response.variableBindings):
            raise ImproperResponse(response.variableBindings)

    def toString(self, depth = 0, tab = "    "):
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            "{}{}:",
            "{}Request ID: {}",
            "{}Error Status: {}",
            "{}Error Index: {}",
            "{}Variable Bindings:",
            "{}"
        )).format(
            indent, typename(self),
            subindent, self.requestID,
            subindent, self.errorStatus.name,
            subindent, self.errorIndex,
            subindent, self.variableBindings.toString(subindent + tab)
        )

    def withRequestID(self, requestID):
        return self.__class__(
            requestID=requestID,
            errorStatus=self.errorStatus,
            errorIndex=self.errorIndex,
            variableBindings=self.variableBindings,
        )

    @classmethod
    def deserialize(cls, data):
        _requestID, errorStatusData     = Integer.decode(data)
        _errorStatus, errorIndexData    = Integer.decode(errorStatusData)
        _errorIndex, data               = Integer.decode(errorIndexData)
        variableBindings                = VarBindList.decodeExact(data)

        requestID = _requestID.value
        errorStatus = _errorStatus.value
        errorIndex = _errorIndex.value

        try:
            errorStatus = ErrorStatus(errorStatus)
        except ValueError as err:
            msg = f"Invalid errorStatus: {errorStatus}"
            raise ParseError(msg, errorStatusData, errorIndexData) from err

        if (errorStatus != ErrorStatus.noError
        and (errorIndex < 0 or errorIndex > len(variableBindings))):
            msg = f"Error index {errorIndex} not valid" \
                f" with {len(variableBindings)} variable bindings"
            raise ParseError(msg, errorIndexData, data)

        return cls(
            requestID=requestID,
            errorStatus=errorStatus,
            errorIndex=errorIndex,
            variableBindings=variableBindings,
        )

class BulkPDU(Constructed):
    READ_CLASS = False
    WRITE_CLASS = False
    RESPONSE_CLASS = False
    NOTIFICATION_CLASS = False
    INTERNAL_CLASS = False
    CONFIRMED_CLASS = False

    def __init__(self,
        *args,
        requestID = 0,
        nonRepeaters = 0,
        maxRepetitions = 1,
        variableBindings = None,
    ):
        if nonRepeaters < 0:
            errmsg = f"nonRepeaters may not be less than 0: {nonRepeaters}"
            raise ValueError(errmsg)

        if maxRepetitions < 0:
            errmsg = f"maxRepetitions may not be less than 0: {maxRepetitions}"
            raise ValueError(errmsg)

        self.requestID = requestID
        self.nonRepeaters = nonRepeaters
        self.maxRepetitions = maxRepetitions

        if variableBindings is None:
            self.variableBindings = VarBindList(*args)
        else:
            self.variableBindings = variableBindings

        if self.nonRepeaters > len(self.variableBindings):
            errmsg = f"The nonRepeaters parameter ({self.nonRepeaters})" \
                " must not exceed the number of variable bindings" \
                f" in the request ({len(self.variableBindings)})"
            raise ValueError(errmsg)

    def __iter__(self):
        yield Integer(self.requestID)
        yield Integer(self.nonRepeaters)
        yield Integer(self.maxRepetitions)
        yield self.variableBindings

    def __len__(self):
        return 4

    def __repr__(self):
        arguments = []
        if self.requestID:
            arguments.append(f"requestID={self.requestID}")

        if self.nonRepeaters:
            arguments.append(f"nonRepeaters={self.nonRepeaters}")

        if self.maxRepetitions:
            arguments.append(f"maxRepetitions={self.maxRepetitions}")

        arguments.append(f"variableBindings={self.variableBindings!r}")

        args = ", ".join(arguments)
        return f"{typename(self)}({args})"

    def __str__(self):
        return self.toString()

    def checkResponse(self, response):
        if not self.validResponse(response.variableBindings):
            raise ImproperResponse(response.variableBindings)

    def toString(self, depth = 0, tab = "    "):
        indent = tab * (depth + 1)
        return "\n".join((
            "{}:",
            "{}Request ID: {}",
            "{}Non-Repeaters: {}",
            "{}Max Repetitions: {}",
            "{}Variable Bindings:",
            "{}"
        )).format(
            typename(self),
            indent, self.requestID,
            indent, self.nonRepeaters,
            indent, self.maxRepetitions,
            indent, self.variableBindings.toString(tab * (depth + 2))
        )

    def withRequestID(self, requestID):
        return self.__class__(
            requestID=requestID,
            nonRepeaters=self.nonRepeaters,
            maxRepetitions=self.maxRepetitions,
            variableBindings=self.variableBindings,
        )

    @classmethod
    def deserialize(cls, data):
        requestID, nrdata = Integer.decode(data)
        nonRepeaters, mrdata = Integer.decode(nrdata)
        maxRepetitions, vbdata = Integer.decode(mrdata)
        variableBindings = VarBindList.decodeExact(vbdata)

        if nonRepeaters.value < 0:
            msg = "nonRepeaters may not be less than 0"
            raise ParseError(msg, nrdata, mrdata)
        elif maxRepetitions.value < 0:
            msg = "maxRepetitions may not be less than 0"
            raise ParseError(msg, mrdata, vbdata)

        return cls(
            requestID=requestID.value,
            nonRepeaters=nonRepeaters.value,
            maxRepetitions=maxRepetitions.value,
            variableBindings=variableBindings,
        )

class GetRequestPDU(PDU):
    CONFIRMED_CLASS = True
    READ_CLASS = True
    TAG = Tag(0, True, Tag.Class.CONTEXT_SPECIFIC)

    def validResponse(self, vblist):
        if len(vblist) != len(self.variableBindings):
            return False

        for request_vb, response_vb in zip(self.variableBindings, vblist):
            if response_vb.name != request_vb.name:
                return False

        return True

class GetNextRequestPDU(PDU):
    CONFIRMED_CLASS = True
    READ_CLASS = True
    TAG = Tag(1, True, Tag.Class.CONTEXT_SPECIFIC)

    def validResponse(self, vblist):
        if len(vblist) != len(self.variableBindings):
            return False

        for request_vb, response_vb in zip(self.variableBindings, vblist):
            if response_vb.name < request_vb.name:
                return False
            elif (response_vb.name == request_vb.name
            and response_vb.value != EndOfMibView()):
                return False

        return True

class ResponsePDU(PDU):
    RESPONSE_CLASS = True
    TAG = Tag(2, True, Tag.Class.CONTEXT_SPECIFIC)

    def checkErrorStatus(self, request):
        if self.errorStatus == ErrorStatus.noError:
            return
        elif self.errorStatus == ErrorStatus.noSuchName:
            raise NoSuchName(self.errorIndex, request)

        raise ErrorResponse(self.errorStatus, self.errorIndex, request)

class SetRequestPDU(PDU):
    CONFIRMED_CLASS = True
    WRITE_CLASS = True
    TAG = Tag(3, True, Tag.Class.CONTEXT_SPECIFIC)

    def __init__(self, *varbinds, **kwargs):
        if kwargs.get("variableBindings") is None:
            vblist = VarBindList(*varbinds, unpack=True)
            kwargs["variableBindings"] = vblist

        super().__init__(**kwargs)

    def validResponse(self, vblist):
        if len(vblist) != len(self.variableBindings):
            return False

        for request_vb, response_vb in zip(self.variableBindings, vblist):
            if response_vb.name != request_vb.name:
                return False

        return True

class GetBulkRequestPDU(BulkPDU):
    CONFIRMED_CLASS = True
    READ_CLASS = True
    TAG = Tag(5, True, Tag.Class.CONTEXT_SPECIFIC)

    def validResponse(self, vblist):
        if len(vblist) < self.nonRepeaters:
            return False

        for i in range(self.nonRepeaters):
            request_vb = self.variableBindings[i]
            response_vb = vblist[i]

            if response_vb.name < request_vb.name:
                return False
            elif (response_vb.name == request_vb.name
            and response_vb.value != EndOfMibView()):
                return False

        repeaters = self.variableBindings[self.nonRepeaters:]
        m = len(repeaters)

        if m > 0:
            repetitions, leftovers = divmod(len(vblist) - self.nonRepeaters, m)

            if repetitions < min(self.maxRepetitions, 1):
                return False
            elif repetitions > self.maxRepetitions:
                return False
            elif leftovers != 0:
                return False

            prev = repeaters
            for r in range(repetitions):
                start = r * m + self.nonRepeaters
                stop = start + m
                successor = vblist[start:stop]

                for i in range(m):
                    request_vb = prev[i]
                    response_vb = successor[i]

                    if response_vb.name < request_vb.name:
                        return False
                    elif (response_vb.name == request_vb.name
                    and response_vb.value != EndOfMibView()):
                        return False

                prev = successor

        return True

class InformRequestPDU(PDU):
    CONFIRMED_CLASS = True
    NOTIFICATION_CLASS = True
    TAG = Tag(6, True, Tag.Class.CONTEXT_SPECIFIC)

class SNMPv2TrapPDU(PDU):
    NOTIFICATION_CLASS = True
    TAG = Tag(7, True, Tag.Class.CONTEXT_SPECIFIC)

class ReportPDU(PDU):
    INTERNAL_CLASS = True
    RESPONSE_CLASS = True
    TAG = Tag(8, True, Tag.Class.CONTEXT_SPECIFIC)

class ErrorResponse(SNMPException):
    def __init__(self, status, index, request):
        self.status = status
        self.index = min(index, len(request.variableBindings))
        self.variableBindings = request.variableBindings

        if self.index == 0:
            self.oid = None
            details = f"{os.linesep}{request}"
        else:
            self.oid = self.variableBindings[self.index-1].name
            details = f" {self.oid}"

        super().__init__(f"{status.name}:{details}")

class NoSuchName(ErrorResponse):
    def __init__(self, index, request):
        super().__init__(ErrorStatus.noSuchName, index, request)

class ImproperResponse(SNMPException):
    def __init__(self, variableBindings):
        self.variableBindings = variableBindings



__all__ = ["Scheduler", "SchedulerTask"]

import heapq
import time

class SchedulerTask:
    def run(self):
        raise NotImplementedError()

class SchedulerEntry:
    def __init__(self, task, timestamp, period):
        if period is not None and period <= 0.0:
            raise ValueError("A repeated task must have a positive period")

        self.task = task
        self.timestamp = timestamp
        self.period = period

    def __call__(self):
        nextTask = self.task.run()

        if nextTask is None or self.period is None:
            return None

        nextTime = self.timestamp + self.period
        return self.__class__(nextTask, nextTime, self.period)

    def __lt__(self, other):
        return self.timestamp < other.timestamp

    def ready(self, timestamp):
        return self.timeToReady(timestamp) <= 0

    def timeToReady(self, timestamp):
        return self.timestamp - timestamp

class Scheduler:
    class ReEntrancyLock:
        def __init__(self):
            self.locked = False

        def __enter__(self):
            self.locked = True

        def __exit__(self, *args, **kwargs):
            self.locked = False

    def __init__(self, sleep_function = time.sleep, time_function = time.time):
        self.lock = self.ReEntrancyLock()
        self.sleep = sleep_function
        self.time = time_function
        self.upcoming = []

    def runPendingTasks(self):
        if self.lock.locked:
            return

        with self.lock:
            while self.upcoming and self.upcoming[0].ready(self.time()):
                nextEntry = self.upcoming[0]()

                if nextEntry is None:
                    heapq.heappop(self.upcoming)
                else:
                    heapq.heapreplace(self.upcoming, nextEntry)

    def schedule(self, task, delay = 0.0, period = None):
        now = self.time()
        entry = SchedulerEntry(task, now + delay, period)
        heapq.heappush(self.upcoming, entry)
        self.runPendingTasks()

    def wait(self):
        if self.upcoming:
            now = self.time()
            entry = self.upcoming[0]
            waitTime = max(entry.timeToReady(now), 0.0)
        else:
            waitTime = 0.0

        self.sleep(waitTime)
        self.runPendingTasks()



__all__ = ["AesCfb128"]

import os

from snmp.smi import OID
from snmp.security.usm import PrivProtocol

from . import *

class AesCfb128(PrivProtocol):
    BYTEORDER = "big"
    CIPHER = AES_128_CFB128

    INTSIZE = 4
    KEYLEN = CIPHER.BLOCKLEN
    SALTLEN = CIPHER.BLOCKLEN - (2 * INTSIZE)
    SALTWRAP = 1 << (8 * SALTLEN)

    def __init__(self, key):
        if len(key) < self.KEYLEN:
            raise ValueError(f"key must be at least {self.KEYLEN} bytes long")

        self.algorithm = OID.parse("1.3.6.1.6.3.10.1.2.4")
        self.key = key[:self.KEYLEN]
        self.salt = int.from_bytes(os.urandom(self.SALTLEN), self.BYTEORDER)

    def __eq__(self, other):
        try:
            return self.algorithm == other.algorithm and self.key == other.key
        except AttributeError:
            return NotImplemented

    def packIV(self, engineBoots, engineTime, salt):
        if len(salt) != self.SALTLEN:
            raise ValueError("Invalid salt")

        return b''.join((
            engineBoots.to_bytes(self.INTSIZE, self.BYTEORDER),
            engineTime .to_bytes(self.INTSIZE, self.BYTEORDER),
            salt
        ))

    def decrypt(self, data, engineBoots, engineTime, salt):
        iv = self.packIV(engineBoots, engineTime, salt)
        return Decryptor(self.CIPHER).decrypt(data, self.key, iv)

    def encrypt(self, data, engineBoots, engineTime):
        self.salt = (self.salt + 1) % self.SALTWRAP

        salt = self.salt.to_bytes(self.SALTLEN, self.BYTEORDER)
        iv = self.packIV(engineBoots, engineTime, salt)
        ciphertext = Encryptor(self.CIPHER).encrypt(data, self.key, iv)

        return ciphertext, salt



from importlib import import_module

packages = (
    "openssl",
    "pycryptodome",
)

modules = (
    "aes",
    "des",
)

for package in packages:
    variables = {}

    for module_name in modules:
        try:
            module = import_module(f".{module_name}", f"{__name__}.{package}")
        except ImportError:
            break

        for variable_name in module.__all__:
            variables[variable_name] = getattr(module, variable_name)

    else:
        globals().update(variables)
        __all__ = list(variables)
        break
else:
    raise ImportError(
        f"{__name__} did not find any supported encryption library"
    )



__all__ = ["MessageIDAuthority", "SNMPv3RequestHandle"]

import weakref

from snmp.exception import *
from snmp.numbers import *
from snmp.requests import *

class MessageIDAuthority(NumberAuthority):
    def newGenerator(self):
        return NumberGenerator(31, signed=False)

    class MessageIDAllocationFailure(SNMPLibraryBug):
        def __init__(self, attempts):
            errmsg = f"No available message ID found after {attempts} attempts"
            super().__init__(errmsg)

    class MessageIDDeallocationFailure(SNMPLibraryBug):
        def __init__(self, msgID: int):
            errmsg = f"Failed to release message ID {msgID}" \
                " because it is not currently reserved"
            super().__init__(errmsg)

    AllocationFailure = MessageIDAllocationFailure
    DeallocationFailure = MessageIDDeallocationFailure

class SNMPv3RequestHandle:
    def __init__(self, scheduler, pdu, *callbacks):
        self.callbacks = list(callbacks)
        self.pdu = pdu
        self.scheduler = scheduler

        self.response = None
        self.exception = None
        self.expired = False

    def __del__(self):
        if self.active():
            self.onDeactivate()

    @property
    def requestID(self):
        return self.pdu.requestID

    def addCallback(self, callback):
        if self.active():
            self.callbacks.append(callback)
        else:
            callback(self.requestID)

    def onDeactivate(self):
        while self.callbacks:
            callback = self.callbacks.pop()
            callback(self.requestID)

    def active(self):
        return self.response is None and not self.expired

    def expire(self):
        if self.active():
            self.expired = True
            self.onDeactivate()

    def push(self, response):
        if self.active():
            self.response = response
            assert self.response is not None
            self.onDeactivate()

    def report(self, exception):
        if self.active() and self.exception is None:
            self.exception = exception

    def wait(self):
        while self.active():
            try:
                self.scheduler.wait()
            except KeyboardInterrupt as interrupt:
                if self.exception is not None:
                    raise self.exception from interrupt
                else:
                    raise

        if self.response is not None:
            self.response.checkErrorStatus(self.pdu)
            self.pdu.checkResponse(self.response)
            return self.response.variableBindings
        elif self.exception is not None:
            raise self.exception
        else:
            assert self.expired
            raise Timeout()



__all__ = ["INTEGER", "OCTET_STRING", "NULL", "OBJECT_IDENTIFIER", "SEQUENCE"]

from snmp.ber import *
from snmp.exception import *
from snmp.utils import *

class ASN1:
    class DeserializeError(SNMPException):
        def __init__(self, msg, etype = ParseError):
            self.etype = etype
            self.msg = msg

        def reraise(self, data, tail = None):
            raise self.etype(self.msg, data, tail) from self

    @classmethod
    def checkTag(cls, tag, data, tail = None):
        if tag != cls.TAG:
            errmsg = f"{tag} does not match the expected type: {cls.TAG}"
            raise ParseError(errmsg, data, tail)

    @classmethod
    def decode(cls, data, **kwargs):
        tag, body, tail = decode(data)
        cls.checkTag(tag, data, tail)

        try:
            return cls.deserialize(body, **kwargs), tail
        except ASN1.DeserializeError as err:
            err.reraise(data, tail)

    @classmethod
    def decodeExact(cls, data, **kwargs):
        tag, body = decodeExact(data)
        cls.checkTag(tag, data)

        try:
            return cls.deserialize(body, **kwargs)
        except ASN1.DeserializeError as err:
            err.reraise(data)

    def encode(self):
        return encode(self.TAG, self.serialize())

    @classmethod
    def deserialize(cls, data):
        raise NotImplementedError()

    def serialize(self):
        raise NotImplementedError()

class Constructed(ASN1):
    def __eq__(self, other):
        if not isinstance(other, Constructed):
            return NotImplemented

        if type(self) != type(other) or len(self) != len(other):
            return False

        for left, right in zip(self, other):
            if left != right:
                return False

        return True

    def __iter__(self):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def serialize(self):
        return b"".join([obj.encode() for obj in self])

class Primitive(ASN1):
    def asOID(self, implied = False):
        raise NotImplementedError()

    @classmethod
    def fromOID(cls, nums, implied = False):
        raise NotImplementedError()

class INTEGER(Primitive):
    BYTEORDER = "big"
    TAG = Tag(2)

    def __init__(self, value):
        self._value = value

    def __eq__(self, other):
        if not isinstance(other, INTEGER):
            return NotImplemented

        return self.value == other.value and self.TAG == other.TAG

    def __repr__(self):
        return f"{typename(self)}({self.value})"

    @property
    def value(self):
        return self._value

    @staticmethod
    def bitCount(value):
        if value < 0:
            value = -value - 1

        return value.bit_length()

    @classmethod
    def construct(cls, value):
        return cls(value)

    @classmethod
    def deserialize(cls, data):
        value = int.from_bytes(data, cls.BYTEORDER, signed=True)
        return cls.construct(value)

    def serialize(self):
        # equivalent to (N + 8) // 8
        # the reason it's not (N + 7) is that ASN.1 always includes a sign bit
        nbytes = (self.bitCount(self.value) // 8) + 1
        return self.value.to_bytes(nbytes, self.BYTEORDER, signed=True)

    def asOID(self, implied = False):
        yield self.value

    @classmethod
    def fromOID(cls, nums, implied = False):
        return cls(next(nums))

class OCTET_STRING(Primitive):
    TAG = Tag(4)

    def __init__(self, data = b""):
        self._data = data

    def __eq__(self, other):
        if not isinstance(other, OCTET_STRING):
            return NotImplemented

        return self.data == other.data and self.TAG == other.TAG

    def __repr__(self):
        return f"{typename(self)}({repr(self.data)})"

    @property
    def data(self):
        return self._data

    def asOID(self, implied = False):
        if not implied:
            yield len(self.data)

        yield from self.data

    @classmethod
    def fromOID(cls, nums, implied = False):
        if implied:
            data = bytes(nums)
        else:
            length = next(nums)
            data = bytes([next(nums) for _ in range(length)])

        return cls(data)

    @classmethod
    def construct(cls, data):
        return cls(data[:])

    @classmethod
    def deserialize(cls, data, copy = True):
        if copy:
            return cls.construct(data[:])
        else:
            return cls.construct(data)

    def serialize(self):
        return self.data

class NULL(Primitive):
    TAG = Tag(5)

    def __eq__(self, other):
        if not isinstance(other, NULL):
            return NotImplemented

        return self.TAG == other.TAG

    def __repr__(self):
        return f"{typename(self)}()"

    def asOID(self, implied = False):
        return ()

    @classmethod
    def fromOID(cls, nums, implied = False):
        return cls()

    @classmethod
    def deserialize(cls, data):
        return cls()

    def serialize(self):
        return b""

class OBJECT_IDENTIFIER(Primitive):
    TAG = Tag(6)

    class BadPrefix(SNMPException):
        pass

    class IndexDecodeError(SNMPException):
        pass

    class CountingIterator:
        def __init__(self, nums, count = 0):
            self.count = count
            self.wrapped = iter(nums[count:])

        def __iter__(self):
            return self

        def __next__(self):
            item = next(self.wrapped)
            self.count += 1
            return item

    def __init__(self, *subidentifiers):
        first = 0
        second = 0

        try:
            first = subidentifiers[0]
            second = subidentifiers[1]
        except IndexError:
            pass

        if first > 2:
            raise ValueError("The first subidentifier must be less than 3")
        elif second >= 40:
            raise ValueError("The second subidentifier must be less than 40")

        if any(map(lambda x: x < 0, subidentifiers)):
            raise ValueError("Sub-identifiers may not be negative")

        self.subidentifiers = subidentifiers

    def __eq__(self, other):
        if not isinstance(other, OBJECT_IDENTIFIER):
            return NotImplemented

        return self.subidentifiers == other.subidentifiers

    def __lt__(self, other):
        return self.subidentifiers < other.subidentifiers

    def __getitem__(self, index):
        return self.subidentifiers.__getitem__(index)

    def __hash__(self):
        return self.subidentifiers.__hash__()

    def __iter__(self):
        return iter(self.subidentifiers)

    def __len__(self):
        return len(self.subidentifiers)

    def __repr__(self):
        args = ", ".join(str(i) for i in self.subidentifiers)
        return f"{typename(self)}({args})"

    def __str__(self):
        return ".".join(str(i) for i in  self.subidentifiers)

    @classmethod
    def parse(cls, oid):
        numbers = oid.split(".")

        if numbers[0] == "":
            numbers = numbers[1:]

        try:
            return cls(*(int(n) for n in numbers))
        except ValueError as err:
            raise ValueError(f"\"{oid}\": {err}") from err

    def asOID(self, implied = False):
        if not implied:
            yield len(self.subidentifiers)

        yield from self.subidentifiers

    @classmethod
    def fromOID(cls, nums, implied = False):
        if implied:
            subidentifiers = nums
        else:
            length = next(nums)
            subidentifiers = [next(nums) for i in range(length)]

        return cls(*subidentifiers)

    def extend(self, *subidentifiers):
        return self.__class__(*self.subidentifiers, *subidentifiers)

    def startswith(self, prefix):
        return self.subidentifiers[:len(prefix)] == prefix[:]

    def decodeIndex(self, prefix_or_position, *types, implied = False):
        if isinstance(prefix_or_position, int):
            position = prefix_or_position
        else:
            prefix = prefix_or_position

            if not self.startswith(prefix):
                errmsg = f"\"{self}\" does not begin with \"{prefix}\""
                raise self.BadPrefix(errmsg)

            position = len(prefix)

        nums = self.CountingIterator(self.subidentifiers, position)

        index = []
        for cls in types[:-1]:
            index.append(self.decodeIndexField(nums, cls))

        for cls in types[-1:]:
            index.append(self.decodeIndexField(nums, cls, implied=implied))

        try:
            next(nums)
        except StopIteration:
            pass
        else:
            errmsg = f"Only {nums.count-1} of {len(self)}" \
                f" sub-identifiers were consumed: \"{self}\""
            raise self.IndexDecodeError(errmsg)

        return tuple(index)

    def decodeIndexField(self, nums, cls, implied = False):
        position = nums.count

        try:
            return cls.fromOID(nums, implied=implied)
        except StopIteration as err:
            impstr = "IMPLIED " if implied else ""
            errmsg = f"Incomplete {impstr}{typename(cls)}" \
                f" at position {position}: {self}"
            raise OBJECT_IDENTIFIER.IndexDecodeError(errmsg) from err
        except ValueError as err:
            impstr = "IMPLIED " if implied else ""
            errmsg = f"Failed to decode {impstr}{typename(cls)}" \
                f" at position {position}: {self}: {err.args[0]}"
            raise OBJECT_IDENTIFIER.IndexDecodeError(errmsg) from err

    def withIndex(self, *index, implied = False):
        oid = self

        for obj in index[:-1]:
            oid = oid.extend(*obj.asOID())

        for obj in index[-1:]:
            oid = oid.extend(*obj.asOID(implied=implied))

        return oid

    @classmethod
    def construct(cls, *subidentifiers):
        try:
            return cls(*subidentifiers)
        except ValueError as err:
            raise ASN1.DeserializeError(err.args[0]) from err

    @classmethod
    def deserialize(cls, data):
        stream = iter(data)

        try:
            oid = list(divmod(next(stream), 40))
        except StopIteration as err:
            raise ASN1.DeserializeError("Empty") from err

        value = 0
        for byte in stream:
            value |= byte & 0x7f
            if byte & 0x80:
                value <<= 7
            else:
                oid.append(value)
                value = 0

        if value:
            errmsg = f"{typename(cls)} encoding ended mid-subidentifier"
            raise ASN1.DeserializeError(errmsg)

        return cls.construct(*oid)

    @staticmethod
    def serializeSubIdentifier(bytearr, num):
        if num < 0x80:
            bytearr.append(num)
        else:
            flag = 0
            tmp = bytearray()

            while num:
                tmp.append((num & 0x7f) | flag)
                flag = 0x80
                num >>= 7

            tmp.reverse()
            bytearr.extend(tmp)

    def serialize(self):
        try:
            first = self.subidentifiers[0]
        except IndexError:
            return b"\x00"

        try:
            second = self.subidentifiers[1]
        except IndexError:
            second = 0

        encoding = bytearray()
        self.serializeSubIdentifier(encoding, first * 40 | second)
        for subidentifier in self.subidentifiers[2:]:
            self.serializeSubIdentifier(encoding, subidentifier)

        return bytes(encoding)

class SEQUENCE(Constructed):
    TAG = Tag(16, True)



__all__ = ["SNMPv1RequestAdmin"]

import weakref

from snmp.exception import *
from snmp.message import *
from snmp.pdu import *
from snmp.requests import *
from snmp.scheduler import *

pduTypes = {
    cls.TAG: cls for cls in (
        GetRequestPDU,
        GetNextRequestPDU,
        ResponsePDU,
        SetRequestPDU,
    )
}

class SNMPv1RequestHandle:
    def __init__(self, scheduler, request):
        self.scheduler = scheduler
        self.callbacks = []

        self.request = request
        self.response = None

        self._expired = False

    def __del__(self):
        if self.active():
            self.close()

    def active(self):
        return self.response is None and not self.expired

    @property
    def expired(self):
        return self._expired

    @expired.setter
    def expired(self, expired):
        if expired and not self._expired:
            self._expired = True
            self.close()

    def close(self):
        while self.callbacks:
            callback, requestID = self.callbacks.pop()
            callback(requestID)

    def addCallback(self, callback, requestID):
        if self.active():
            self.callbacks.append((callback, requestID))
        else:
            callback(requestID)

    def push(self, message):
        if self.active():
            self.response = message.pdu
            self.close()

    def wait(self):
        while self.active():
            self.scheduler.wait()

        if self.response is not None:
            self.response.checkErrorStatus(self.request)
            self.request.checkResponse(self.response)
            return self.response.variableBindings
        else:
            raise Timeout()

class SNMPv1RequestAdmin:
    class ExpireTask(SchedulerTask):
        def __init__(self, handle_ref):
            self.handle_ref = handle_ref

        def run(self):
            handle = self.handle_ref()
            if handle is not None:
                handle.expired = True

    class SendTask(SchedulerTask):
        def __init__(self, handle_ref, channel, community):
            self.handle_ref = handle_ref
            self.channel = channel

            handle = self.handle_ref()
            if handle is not None:
                self.msg = Message(
                    ProtocolVersion.SNMPv1,
                    community,
                    handle.request,
                ).encode()

        def run(self):
            handle = self.handle_ref()
            if handle is not None and handle.active():
                self.channel.send(self.msg)
                return self

    def __init__(self, scheduler):
        self.requestIDAuthority = RequestIDAuthority()
        self.scheduler = scheduler
        self.outstanding = {}

    def closeRequest(self, requestID):
        try:
            del self.outstanding[requestID]
        except KeyError:
            pass

        self.requestIDAuthority.release(requestID)

    def openRequest(self, pdu, community, channel, timeout, refreshPeriod):
        requestID = self.requestIDAuthority.reserve()
        request = pdu.withRequestID(requestID)

        handle = SNMPv1RequestHandle(self.scheduler, request)
        reference = weakref.ref(handle)
        self.outstanding[requestID] = reference, community
        handle.addCallback(self.closeRequest, requestID)

        expireTask = self.ExpireTask(reference)
        self.scheduler.schedule(expireTask, timeout)

        sendTask = self.SendTask(reference, channel, community)
        self.scheduler.schedule(sendTask, period=refreshPeriod)

        return handle

    def hear(self, data, channel):
        message = Message.decodeExact(data, types=pduTypes)

        try:
            reference, community = self.outstanding[message.pdu.requestID]
        except KeyError as err:
            errmsg = f"Unknown requestID: {message.pdu.requestID}"
            raise IncomingMessageError(errmsg)

        handle = reference()

        if handle is None:
            errmsg = f"Request {message.pdu.requestID} was not properly closed"
            raise SNMPLibraryBug(errmsg)

        if message.community != community:
            errmsg = "Wrong community name in response to" \
                f" request {message.pdu.requestID}:" \
                f" {message.community!r} != {community!r}"
            raise IncomingMessageError(errmsg)

        handle.push(message)



__all__ = ["SecurityLevel", "noAuthNoPriv", "authNoPriv", "authPriv"]

from snmp.utils import typename

class SecurityLevel:
    def __init__(self, auth = False, priv = False):
        a = bool(auth)
        p = bool(priv)

        if p and not a:
            raise ValueError("Privacy without authentication is not valid")

        self._auth = a
        self._priv = p

    def __repr__(self):
        return "{}(auth={}, priv={})".format(
            typename(self),
            self.auth,
            self.priv
        )

    def __str__(self):
        return "{}{}".format(
            "auth" if self.auth else "noAuth",
            "Priv" if self.priv else "NoPriv"
        )

    @property
    def auth(self):
        return self._auth

    @property
    def priv(self):
        return self._priv

    def __eq__(self, other):
        try:
            result = (self.auth == other.auth and self.priv == other.priv)
        except AttributeError:
            return NotImplemented
        else:
            return result

    def __lt__(self, other):
        if self.auth:
            return other.priv and not self.priv
        else:
            return other.auth

    def __ge__(self, other):
        return not self < other

noAuthNoPriv = SecurityLevel()
authNoPriv = SecurityLevel(auth=True)
authPriv = SecurityLevel(auth=True, priv=True)



__all__ = [
    "HeaderData", "MessageFlags", "ScopedPDU", "SecurityName",
    "SNMPv3Message", "SNMPv3WireMessage", "ReportMessage",
]

from snmp.exception import *
from snmp.ber import *
from snmp.message import InvalidMessage, ProtocolVersion
from snmp.pdu import *
from snmp.smi import *
from snmp.utils import *

from snmp.security import *
from snmp.security.levels import *

pduTypes = {
    cls.TAG: cls for cls in (
        GetRequestPDU,
        GetNextRequestPDU,
        ResponsePDU,
        SetRequestPDU,
        GetBulkRequestPDU,
        InformRequestPDU,
        SNMPv2TrapPDU,
        ReportPDU,
    )
}

class SecurityName:
    def __init__(self, userName, *namespaces):
        self.userName = userName
        self.namespaces = set(namespaces)

    def __eq__(self, other):
        try:
            return (self.userName == other.userName
            and self.namespaces == other.namespaces)
        except AttributeError:
            return NotImplemented

    def __repr__(self):
        args = (repr(self.userName), *map(repr, self.namespaces))
        return f"{typename(self)}({', '.join(args)})"

class MessageFlags(OctetString):
    AUTH_FLAG        = (1 << 0)
    PRIV_FLAG        = (1 << 1)
    REPORTABLE_FLAG  = (1 << 2)

    def __init__(self, securityLevel = noAuthNoPriv, reportable = False):
        self.securityLevel = securityLevel
        self.reportableFlag = reportable

    def __repr__(self):
        return f"{typename(self)}({self.securityLevel}, {self.reportableFlag})"

    def __str__(self):
        return self.toString()

    def toString(self, depth = 0, tab = "    "):
        indent = tab * depth
        subindent = indent + tab

        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Security Level: {self.securityLevel}",
            f"{subindent}Reportable: {self.reportableFlag}",
        ))

    @classmethod
    def construct(cls, data):
        try:
            byte = data[0]
        except IndexError as err:
            errmsg = f"{typename(cls)} must contain at least one byte"
            raise cls.DeserializeError(errmsg) from err

        try:
            securityLevel = SecurityLevel(
                byte & cls.AUTH_FLAG,
                byte & cls.PRIV_FLAG,
            )
        except ValueError as err:
            errmsg = f"Invalid msgFlags: {err.args[0]}"
            raise cls.DeserializeError(errmsg, InvalidMessage) from err

        reportable = (byte & cls.REPORTABLE_FLAG != 0)
        return cls(securityLevel, reportable)

    @property
    def data(self):
        byte = 0

        if self.authFlag:
            byte |= self.AUTH_FLAG

        if self.privFlag:
            byte |= self.PRIV_FLAG

        if self.reportableFlag:
            byte |= self.REPORTABLE_FLAG

        return bytes((byte,))

    @property
    def authFlag(self):
        return self.securityLevel.auth

    @property
    def privFlag(self):
        return self.securityLevel.priv

class HeaderData(Sequence):
    def __init__(self, msgID, maxSize, flags, securityModel):
        if msgID < 0:
            raise ValueError(f"Message ID may not be negative: {msgID}")

        if maxSize < 484:
            raise ValueError(f"msgMaxSize must be at least 484: {maxSize}")

        self._msgID = Integer(msgID)
        self._maxSize = Integer(maxSize)
        self.flags = flags
        self.securityModel = securityModel

    @property
    def id(self):
        return self.msgID

    @property
    def msgID(self):
        return self._msgID.value

    @property
    def maxSize(self):
        return self._maxSize.value

    def __iter__(self):
        yield self._msgID
        yield self._maxSize
        yield self.flags
        yield Integer(self.securityModel)

    def __len__(self):
        return 4

    def __repr__(self):
        args = (
            str(self.id),
            str(self.maxSize),
            repr(self.flags),
            str(self.securityModel),
        )

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self):
        return self.toString()

    def toString(self, depth = 0, tab = "    "):
        indent = tab * depth
        subindent = indent + tab
        securityModel = self.securityModel

        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Message ID: {self.id}",
            f"{subindent}Sender Message Size Limit: {self.maxSize}",
            f"{self.flags.toString(depth+1, tab)}",
            f"{subindent}Security Model: {securityModel.name}"
        ))

    def withMessageID(self, messageID):
        return HeaderData(
            messageID,
            self.maxSize,
            self.flags,
            self.securityModel,
        )

    @classmethod
    def deserialize(cls, data):
        msgID, msdata = Integer.decode(data)

        if msgID.value < 0:
            errmsg = f"Message ID may not be negative: {msgID.value}"
            raise ParseError(errmsg, subbytes(data), msdata)

        msgMaxSize, mfdata = Integer.decode(msdata)

        if msgMaxSize.value < 484:
            errmsg = f"msgMaxSize must be at least 484: {msgMaxSize.value}"
            raise ParseError(errmsg, msdata, mfdata)

        msgFlags, smdata = MessageFlags.decode(mfdata)
        msgSecurityModel = Integer.decodeExact(smdata)

        if msgSecurityModel.value < 1:
            errmsg = "msgSecurityModel may not be less than 1"
            raise ParseError(errmsg, smdata)

        try:
            securityModel = SecurityModel(msgSecurityModel.value)
        except ValueError as err:
            errmsg = f"Unknown security model: {msgSecurityModel.value}"
            raise UnknownSecurityModel(errmsg, smdata) from err

        try:
            return cls(msgID.value, msgMaxSize.value, msgFlags, securityModel)
        except ValueError as err:
            raise cls.DeserializeError(err.args[0]) from err

class ScopedPDU(Sequence):
    def __init__(self, pdu, contextEngineID, contextName = b""):
        self._contextEngineID = OctetString(contextEngineID)
        self._contextName = OctetString(contextName)
        self.pdu = pdu

    @property
    def contextEngineID(self):
        return self._contextEngineID.data

    @property
    def contextName(self):
        return self._contextName.data

    def __iter__(self):
        yield self._contextEngineID
        yield self._contextName
        yield self.pdu

    def __len__(self):
        return 3

    def __repr__(self):
        args = (
            repr(self.pdu),
            repr(self.contextEngineID),
            f"contextName={repr(self.contextName)}"
        )

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self):
        return self.toString()

    def toString(self, depth = 0, tab = "    "):
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Context Engine ID: {self.contextEngineID!r}",
            f"{subindent}Context Name: {self.contextName!r}",
            f"{self.pdu.toString(depth=depth+1, tab=tab)}"
        ))

    def withContextEngineID(self, engineID):
        return ScopedPDU(self.pdu, engineID, self.contextName)

    @classmethod
    def deserialize(cls, data):
        contextEngineID, data   = OctetString.decode(data)
        contextName, data       = OctetString.decode(data)
        tag, _                  = Tag.decode(data)

        try:
            pduType = pduTypes[tag]
        except KeyError as err:
            errmsg = f"{typename(cls)} does not support PDUs of type {tag}"
            raise ParseError(errmsg, data) from err

        return cls(
            pduType.decodeExact(data),
            contextEngineID = contextEngineID.data,
            contextName     = contextName.data,
        )

class SNMPv3Message:
    def __init__(self, header, scopedPDU, securityEngineID, securityName):
        self.header = header
        self.scopedPDU = scopedPDU
        self.securityEngineID = securityEngineID
        self.securityName = securityName

    def __eq__(self, other):
        try:
            return (self.header == other.header
                and self.scopedPDU == other.scopedPDU
                and self.securityEngineID == other.securityEngineID
                and self.securityName == other.securityName
            )
        except AttributeError:
            return NotImplemented

    def __repr__(self):
        args = (repr(field) for field in (
            self.header,
            self.scopedPDU,
            self.securityEngineID,
            self.securityName,
        ))

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self):
        return self.toString()

    def toString(self, depth = 0, tab = "    "):
        indent = tab * depth
        subindent = indent + tab

        return "\n".join((
            f"{indent}{typename(self)}:",
            self.header.toString(depth+1, tab),
            f"{subindent}Security EngineID: {self.securityEngineID}",
            f"{subindent}Security Name: {self.securityName.userName}",
            self.scopedPDU.toString(depth+1, tab),
        ))

    def withEngineID(self, engineID):
        return SNMPv3Message(
            self.header,
            self.scopedPDU.withContextEngineID(engineID),
            engineID,
            self.securityName,
        )

    def withMessageID(self, messageID):
        return SNMPv3Message(
            self.header.withMessageID(messageID),
            self.scopedPDU,
            self.securityEngineID,
            self.securityName,
        )

class SNMPv3WireMessage(Sequence):
    VERSION = ProtocolVersion.SNMPv3

    def __init__(self, header, scopedPduData, securityParameters):
        if header.flags.privFlag:
            if not isinstance(scopedPduData, OctetString):
                raise TypeError(
                    "scopedPduData must be an OctetString"
                    " when the privFlag is set"
                )
        else:
            if not isinstance(scopedPduData, ScopedPDU):
                raise TypeError(
                    "scopedPduData must be a ScopedPDU"
                    " when the privFlag is unset"
                )

        self.header = header
        self.scopedPduData = scopedPduData
        self.securityParameters = securityParameters

    def __iter__(self):
        yield Integer(self.VERSION)
        yield self.header
        yield self.securityParameters
        yield self.scopedPduData

    def __len__(self):
        return 4;

    def __repr__(self):
        args = (repr(field) for field in (
            self.header,
            self.scopedPduData,
            self.securityParameters,
        ))

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self):
        return self.toString()

    def toString(self, depth = 0, tab = "    "):
        indent = tab * depth
        subindent = indent + tab

        if self.header.flags.privFlag:
            payload = f"{subindent}Encrypted Data: {self.scopedPduData}"
        else:
            payload = self.scopedPduData.toString(depth+1, tab)

        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{self.header.toString(depth+1, tab)}",
            f"{subindent}Security Parameters: {self.securityParameters}",
            payload,
        ))

    @classmethod
    def deserialize(cls, data):
        msgVersion, ptr = Integer.decode(data)

        try:
            version = ProtocolVersion(msgVersion.value)
        except ValueError as err:
            errmsg = f"Unsupported msgVersion: {msgVersion.value}"
            raise ParseError(errmsg, subbytes(data), ptr) from err

        if version != cls.VERSION:
            errmsg = f"{typename(cls)} cannot decode {version.name} messages"
            raise ParseError(errmsg, subbytes(data), ptr)

        msgGlobalData, ptr = HeaderData.decode(ptr)
        msgSecurityData, ptr = OctetString.decode(ptr, copy=False)

        if msgGlobalData.flags.privFlag:
            scopedPduData = OctetString.decodeExact(ptr)
        else:
            scopedPduData = ScopedPDU.decodeExact(ptr)

        return cls(msgGlobalData, scopedPduData, msgSecurityData)

    @classmethod
    def findSecurityParameters(self, wholeMsg):
        tag, ptr, tail      = decode(wholeMsg)
        tag, version, ptr   = decode(ptr)
        tag, header, ptr    = decode(ptr)
        tag, ptr, tail      = decode(ptr)
        return ptr

    @staticmethod
    def decodePlaintext(data):
        scopedPDU, padding = ScopedPDU.decode(data)
        return scopedPDU

class ReportMessage(SNMPException):
    def __init__(self, message):
        self.message = message



__all__ = [
    "AuthenticationNotEnabled", "InvalidSignature", "PrivacyNotEnabled",
    "Credentials", "AuthCredentials", "AuthPrivCredentials",
    "LocalizedCredentials",
]

from snmp.exception import *
from snmp.security.levels import *
from snmp.security.usm.parameters import UnsignedUsmParameters
from snmp.smi import OctetString
from snmp.utils import *
from snmp.v3.message import *

class AuthenticationNotEnabled(SNMPException):
    pass

class PrivacyNotEnabled(SNMPException):
    pass

class InvalidSignature(IncomingMessageError):
    pass

class WrongSignatureLength(InvalidSignature):
    pass

class LocalizedCredentials:
    def __eq__(self, other):
        if type(self) == type(other):
            return True
        else:
            return NotImplemented

    def __repr__(self):
        return f"{typename(self)}()"

    def withoutPrivacy(self):
        return self

    def sign(self, message):
        raise AuthenticationNotEnabled()

    def signaturePlaceholder(self):
        raise AuthenticationNotEnabled()

    def verifySignature(self, signature):
        raise AuthenticationNotEnabled()

    def decrypt(self, encryptedPDU, msgBoots, msgTime, salt):
        raise PrivacyNotEnabled()

    def encrypt(self, scopedPDU, snmpEngineBoots, snmpEngineTime):
        raise PrivacyNotEnabled()

class LocalizedAuthCredentials(LocalizedCredentials):
    def __init__(self, auth):
        super().__init__()
        self.auth = auth

    def __eq__(self, other):
        result = super().__eq__(other)
        if result is True:
            return self.auth == other.auth
        else:
            return result

    def __repr__(self):
        return f"{typename(self)}({self.auth!r})"

    def sign(self, message):
        wholeMsg = message.encode()
        ptr = message.findSecurityParameters(wholeMsg)
        padding = UnsignedUsmParameters.findPadding(ptr)
        signature = self.auth.sign(wholeMsg)
        return padding.replace(signature)

    def signaturePlaceholder(self):
        return self.auth.msgAuthenticationParameters

    def verifySignature(self, signature):
        padding = self.signaturePlaceholder()

        if len(signature) != len(padding):
            raise WrongSignatureLength(signature)

        wholeMsg = signature.replace(padding)
        computed = self.auth.sign(wholeMsg)

        if computed != signature:
            raise InvalidSignature(signature, computed)

class LocalizedAuthPrivCredentials(LocalizedAuthCredentials):
    def __init__(self, auth, priv):
        super().__init__(auth)
        self.priv = priv

    def __eq__(self, other):
        result = super().__eq__(other)
        if result is True:
            return self.priv == other.priv
        else:
            return result

    def __repr__(self):
        return f"{typename(self)}({self.auth!r}, {self.priv!r})"

    def withoutPrivacy(self):
        return LocalizedAuthCredentials(self.auth)

    def decrypt(self, encryptedPDU, msgBoots, msgTime, salt):
        ciphertext = encryptedPDU.data
        plaintext = self.priv.decrypt(ciphertext, msgBoots, msgTime, salt)
        return SNMPv3WireMessage.decodePlaintext(plaintext)

    def encrypt(self, scopedPDU, snmpEngineBoots, snmpEngineTime):
        ciphertext, salt = self.priv.encrypt(
            scopedPDU.encode(),
            snmpEngineBoots,
            snmpEngineTime,
        )

        return OctetString(ciphertext), salt

class Credentials:
    def __init__(self):
        self.maxSecurityLevel = noAuthNoPriv

    def localize(self, engineID):
        return LocalizedCredentials()

class AuthCredentials(Credentials):
    def __init__(self, authProtocol, authSecret):
        super().__init__()
        self.maxSecurityLevel = authNoPriv
        self.authProtocol = authProtocol
        self.authKey = self.authProtocol.computeKey(authSecret)

    def localizeAuth(self, engineID):
        key = self.authProtocol.localizeKey(self.authKey, engineID)
        return self.authProtocol(key)

    def localize(self, engineID):
        return LocalizedAuthCredentials(self.localizeAuth(engineID))

class AuthPrivCredentials(AuthCredentials):
    def __init__(self,
        authProtocol,
        privProtocol,
        authSecret = None,
        privSecret = None,
        secret = None,
    ):
        if secret is None:
            if authSecret is None:
                raise TypeError("missing required argument: 'authSecret'")
            elif privSecret is None:
                raise TypeError("missing required argument: 'privSecret'")

            super().__init__(authProtocol, authSecret)
            self.privKey = self.authProtocol.computeKey(privSecret)
        else:
            if authSecret is not None:
                errmsg = "'authSecret' and 'secret' are mutually exclusive"
                raise TypeError(errmsg)
            elif privSecret is not None:
                errmsg = "'privSecret' and 'secret' are mutually exclusive"
                raise TypeError(errmsg)

            super().__init__(authProtocol, secret)
            self.privKey = self.authKey

        self.maxSecurityLevel = authPriv
        self.privProtocol = privProtocol

    def localizePriv(self, engineID):
        key = self.authProtocol.localizeKey(self.privKey, engineID)
        return self.privProtocol(key)

    def localize(self, engineID):
        auth = self.localizeAuth(engineID)
        priv = self.localizePriv(engineID)
        return LocalizedAuthPrivCredentials(auth, priv)



__all__ = [
    "Engine", "ErrorResponse", "ErrorStatus",
    "ImproperResponse", "NoSuchName", "Timeout",
    "SNMPv1", "SNMPv2c", "SNMPv3",
    "UDP_IPv4", "UDP_IPv6",
    "noAuthNoPriv", "authNoPriv", "authPriv",
    "NoSuchObject", "NoSuchInstance", "EndOfMibView", "OID",
]

from snmp.engine import Engine
from snmp.message import ProtocolVersion
from snmp.requests import Timeout
from snmp.transport import TransportDomain
from snmp.security.levels import noAuthNoPriv, authNoPriv, authPriv
from snmp.smi import OID
from snmp.pdu import (
    EndOfMibView, NoSuchInstance, NoSuchObject,
    ErrorStatus, ErrorResponse, NoSuchName, ImproperResponse,
)

SNMPv1  = ProtocolVersion.SNMPv1
SNMPv2c = ProtocolVersion.SNMPv2c
SNMPv3  = ProtocolVersion.SNMPv3

UDP_IPv4 = TransportDomain.UDP_IPv4
UDP_IPv6 = TransportDomain.UDP_IPv6



__all__ = ["SNMPv1Manager"]

from snmp.pdu import *

class SNMPv1Manager:
    def __init__(self, admin, channel, community, autowait = True):
        self.autowait = autowait
        self.channel = channel

        self.admin = admin
        self.defaultCommunity = community

    def sendRequest(self,
        request,
        community = None,
        wait = None,
        timeout = 10.0,
        refreshPeriod = 1.0,
    ):
        if community is None:
            community = self.defaultCommunity

        if wait is None:
            wait = self.autowait

        handle = self.admin.openRequest(
            request,
            community,
            self.channel,
            timeout,
            refreshPeriod,
        )

        if wait:
            return handle.wait()
        else:
            return handle

    def get(self, *oids, **kwargs):
        pdu = GetRequestPDU(*oids)
        return self.sendRequest(pdu, **kwargs)

    def getBulk(self, *oids, nonRepeaters=0, maxRepetitions=1, **kwargs):
        # Validate arguments
        _ = GetBulkRequestPDU(
            *oids,
            nonRepeaters=nonRepeaters,
            maxRepetitions=maxRepetitions,
        )

        if maxRepetitions == 0:
            oids = oids[:nonRepeaters]

        pdu = GetNextRequestPDU(*oids)
        return self.sendRequest(pdu, **kwargs)

    def getNext(self, *oids, **kwargs):
        pdu = GetNextRequestPDU(*oids)
        return self.sendRequest(pdu, **kwargs)

    def set(self, *varbinds, **kwargs):
        pdu = SetRequestPDU(*varbinds)
        return self.sendRequest(pdu, **kwargs)



__all__ = ["Message"]

from snmp.ber import ParseError, Tag
from snmp.smi import *
from snmp.utils import *

from .version import *

class Message(Sequence):
    VERSIONS = (ProtocolVersion.SNMPv1, ProtocolVersion.SNMPv2c)

    def __init__(self, version, community, pdu):
        self.version = version
        self.community = community
        self.pdu = pdu

    def __iter__(self):
        yield Integer(self.version)
        yield OctetString(self.community)
        yield self.pdu

    def __len__(self):
        return 3

    def __repr__(self):
        return "{}({}, {!r}, {})".format(
            typename(self),
            str(self.version),
            self.community,
            repr(self.pdu),
        )

    def __str__(self):
        return self.toString()

    def toString(self, depth = 0, tab = "    "):
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Community: {self.community!r}",
            f"{self.pdu.toString(depth+1, tab)}",
        ))

    @classmethod
    def deserialize(cls, data, types = None):
        msgVersion, ptr = Integer.decode(data)

        try:
            version = ProtocolVersion(msgVersion.value)
        except ValueError as err:
            errmsg = f"Invalid msgVersion: {msgVersion.value}"
            raise BadVersion(errmsg, data, ptr) from err

        if version not in cls.VERSIONS:
            errmsg = f"{typename(cls)} does not support {version.name}"
            raise BadVersion(errmsg, data, ptr)

        community, ptr = OctetString.decode(ptr)
        tag, _ = Tag.decode(subbytes(ptr))

        if types is None:
            types = dict()

        try:
            pduType = types[tag]
        except KeyError as err:
            raise ParseError(f"Invalid PDU type: {tag}", ptr) from err

        return cls(
            version,
            community.data,
            pduType.decodeExact(ptr),
        )



__all__ = ["UdpIPv4Socket", "UdpIPv6Socket", "UdpMultiplexor"]

from importlib import import_module
from socket import *

from snmp.transport import *
from snmp.transport import package

class UdpSocket:
    DEFAULT_PORT = {
        AddressUsage.LISTENER: 161,
        AddressUsage.SENDER: 0,
        AddressUsage.TRAP_LISTENER: 162,
    }

    @classmethod
    def normalizeAddress(cls, address = None, usage = AddressUsage.SENDER):
        if isinstance(address, tuple):
            addr, port = address
        else:
            if address is None:
                addr = cls.DOMAIN.default_address
            else:
                addr = address

            port = cls.DEFAULT_PORT[usage]

        permissive = usage not in (
            AddressUsage.LISTENER,
            AddressUsage.TRAP_LISTENER,
        )

        if addr == "" and permissive:
            addr = cls.DOMAIN.default_address

        try:
            _ = inet_pton(cls.DOMAIN.address_family, addr)
        except OSError as err:
            family = cls.DOMAIN.address_family.name
            errstr = f"Invalid address for {family}: \"{addr}\""
            raise ValueError(errstr) from err
        except TypeError as err:
            raise TypeError(f"Address must be a str: {addr}") from err

        if port < 0 or port > 0xffff or (port == 0 and not permissive):
            errmsg = "Invalid UDP port number: {}"
            raise ValueError(errmsg.format(port))

        return addr, port

    @property
    def fileno(self):
        return self.socket.fileno()

    @property
    def port(self):
        return self.socket.getsockname()[1]

    def __init__(self, host, port, recvSize):
        self.recvSize = recvSize

        self.socket = socket(self.DOMAIN.address_family, SOCK_DGRAM)
        self.socket.setblocking(False)
        self.socket.bind((host, port))

    def receive(self):
        data, addr = self.socket.recvfrom(self.recvSize)
        return addr, data

    def send(self, data, address):
        self.socket.sendto(data, address)

    def close(self):
        self.socket.close()

class UdpIPv4Socket(UdpSocket):
    DOMAIN = TransportDomain.UDP_IPv4

    def __init__(self, host = "", port = 0, mtu = 1500):
        super().__init__(host, port, mtu - 28)

class UdpIPv6Socket(UdpSocket):
    DOMAIN = TransportDomain.UDP_IPv6

    def __init__(self, host = "", port = 0, mtu = 1500):
        super().__init__(host, port, mtu - 48)

module = import_module(".udp", package=package)
UdpMultiplexor = module.UdpMultiplexor



__all__ = ["RequestIDAuthority", "Timeout"]

from snmp.exception import *
from snmp.numbers import *

class Timeout(SNMPException):
    pass

class RequestIDAuthority(NumberAuthority):
    class RequestIDAllocationFailure(SNMPLibraryBug):
        # This error is an SNMPLibraryBug type because it should not be
        # possible to run out of available requestIDs. The generator would
        # have had to pass through the full range of requestIDs, and remain
        # so densely full that a new generator could not find an opening. If
        # everything was working properly, there would have to be over a
        # billion outstanding requests in order to ever see this error. On
        # it's own, having a billion outstanding requests would seem to
        # indicate a bug, and, more importantly, encountering this error by
        # any other way would also indicate a bug, so that's why it's a type
        # of SNMPLibraryBug.
        def __init__(self, attempts):
            errmsg = f"No available request ID found after {attempts} attempts"
            super().__init__(errmsg)

            super().__init__(errmsg)

    class RequestIDDeallocationFailure(SNMPLibraryBug):
        def __init__(self, requestID):
            errmsg = f"Failed to release request ID {requestID}" \
                " because it is not currently reserved"
            super().__init__(errmsg)

    AllocationFailure = RequestIDAllocationFailure
    DeallocationFailure = RequestIDDeallocationFailure

    def newGenerator(self) -> NumberGenerator:
        return NumberGenerator(32, signed=True)
